#/usr/bin/env python
"""
_Destroy_

Implementation of DBSBuffer.Destroy for Oracle
"""




import threading

from WMCore.Database.DBCreator import DBCreator

class Destroy(DBCreator):
    def __init__(self):
        """
        _init_

        Call the base class's constructor and create all necessary tables,
        constraints and inserts.
        """
        myThread = threading.currentThread()
        DBCreator.__init__(self, myThread.logger, myThread.dbi)


        self.delete["10dbsbuffer_dataset"]            = "DROP TABLE dbsbuffer_dataset"
        self.delete["09dbsbuffer_algo"]               = "DROP TABLE dbsbuffer_algo"
        self.delete["08dbsbuffer_algo_dataset_assoc"] = "DROP TABLE dbsbuffer_algo_dataset_assoc"
        self.delete["07dbsbuffer_file"]               = "DROP TABLE dbsbuffer_file"
        self.delete["06dbsbuffer_file_parent"]        = "DROP TABLE dbsbuffer_file_parent"
        self.delete["05dbsbuffer_file_runlumi_map"]   = "DROP TABLE dbsbuffer_file_runlumi_map"
        self.delete["12dbsbuffer_location"]           = "DROP TABLE dbsbuffer_location"
        self.delete["03dbsbuffer_file_location"]      = "DROP TABLE dbsbuffer_file_location"
        self.delete["02dbsbuffer_block"]              = "DROP TABLE dbsbuffer_block"
        self.delete["02dbsbuffer_checksum_type"]      = "DROP TABLE dbsbuffer_checksum_type"
        self.delete["01dbsbuffer_file_checksums"]     = "DROP TABLE dbsbuffer_file_checksums"

        self.delete["01dbsbuffer_dataset_seq"]            = "DROP SEQUENCE dbsbuffer_dataset_seq"
        self.delete["02dbsbuffer_algo_seq"]               = "DROP SEQUENCE dbsbuffer_algo_seq"
        self.delete["03dbsbuffer_file_seq"]               = "DROP SEQUENCE dbsbuffer_file_seq"
        self.delete["04dbsbuffer_location_seq"]           = "DROP SEQUENCE dbsbuffer_location_seq"
        self.delete["05dbsbuffer_algo_dataset_assoc_seq"] = "DROP SEQUENCE dbsbuffer_algdset_assoc_seq"
        self.delete["01dbsbuffer_block_seq"]              = "DROP SEQUENCE dbsbuffer_block_seq"
        self.delete["01dbsbuffer_checksum_type_seq"]      = "DROP SEQUENCE dbsbuffer_checksum_type_seq"


        self.delete["01dbsbuffer_dataset_trg"] = "DROP TRIGGER dbsbuffer_dataset_trg"
        self.delete["02dbsbuffer_algo_trg"] = "DROP TRIGGER dbsbuffer_algo_trg"
        self.delete["03dbsbuffer_file_trg"] = "DROP TRIGGER dbsbuffer_file_trg"
        #self.create["04dbsbuffer_location_trg"] = "DROP TRIGGER dbsbuffer_location_trg"
        self.delete["05dbsbuffer_algo_dataset_assoc_trg"] = "DROP TRIGGER dbsbuffer_algdset_assoc_trg"
        #self.delete["06dbsbuffer_block_trg"] = "DROP TRIGGER dbsbuffer_block_trg"



#!/usr/bin/env python
"""
_BulkNewReturn_

MySQL implementation of JobGroup.BulkNewReturn
"""




import time

from WMCore.Database.DBFormatter import DBFormatter

class BulkNewReturn(DBFormatter):
    """
    Does a bulk commit of jobGroups, followed by returning their IDs and UIDs

    """
    sql = """INSERT INTO wmbs_jobgroup (subscription, guid, output,
             last_update) VALUES (:subscription, :guid, :output,
             :timestamp)"""

    returnSQL = """SELECT id AS id, guid AS guid FROM wmbs_jobgroup
                   WHERE subscription = :subscription
                   AND guid = :guid
                   AND output = :output"""

    def execute(self, bulkInput = None, conn = None, transaction = False):
        """
        This can take a list of dictionaries {subscription, guid, output}
        instead of the original inputs

        """
        timestamp = int(time.time())
        insertBinds = []
        returnBinds = []
        for entry in bulkInput:
            insertBinds.append({'subscription': entry['subscription'],
                                'guid': entry['uid'],
                                'output': entry['output'],
                                'timestamp': timestamp})
            returnBinds.append({'subscription': entry['subscription'],
                                'guid': entry['uid'],
                                'output': entry['output']})


        self.dbi.processData(self.sql, insertBinds, conn = conn,
                             transaction = transaction)
        result = self.dbi.processData(self.returnSQL, returnBinds,
                                      conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



"""
_GetHeartbeatInfo_

Fetches the hearbeat info for all worker threads associated to
the component name passed in.
"""

__all__ = []

from WMCore.Database.DBFormatter import DBFormatter


class GetHeartbeatInfo(DBFormatter):
    sql = """SELECT comp.name as name, comp.pid, worker.name as worker_name,
                    worker.state, worker.last_updated, comp.update_threshold,
                    worker.poll_interval, worker.cycle_time, worker.outcome,
                    worker.last_error, worker.error_message
             FROM wm_workers worker
             INNER JOIN wm_components comp ON comp.id = worker.component_id
             WHERE comp.name = :component_name
             ORDER BY worker.last_updated ASC
             """

    def execute(self, compName, conn=None, transaction=False):
        bind = {"component_name": compName}

        result = self.dbi.processData(self.sql, bind, conn=conn,
                                      transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_Subscription.New_

Oracle implementation of Subscription.New
"""

from WMCore.WMBS.MySQL.Subscriptions.New import New as NewMySQL

class New(NewMySQL):
    typesSQL = """INSERT INTO wmbs_sub_types (id, name)
                    SELECT wmbs_sub_types_SEQ.nextval, :subtype FROM dual
                    WHERE NOT EXISTS (SELECT id FROM wmbs_sub_types WHERE name = :subtype)"""

    sql = """INSERT INTO wmbs_subscription (id, fileset, workflow, subtype,
                                            split_algo, last_update)
               SELECT wmbs_subscription_SEQ.nextval, :fileset, :workflow, id,
                      :split_algo, :timestamp FROM wmbs_sub_types
               WHERE name = :subtype"""



"""
Oracle implementation of GetParentsFile
"""

from WMCore.WMBS.MySQL.Files.GetParents import GetParents as GetParentsFileMySQL

class GetParents(GetParentsFileMySQL):
    sql = GetParentsFileMySQL.sql



#!/usr/bin/env python
"""
_SetPendingJobSlotsForSite_

Set the number of pending job slots available at a given site.
"""

from WMCore.Database.DBFormatter import DBFormatter

class SetPendingJobSlotsForSite(DBFormatter):
    sql = "UPDATE wmbs_location SET pending_slots = :slots WHERE site_name = :site"

    def execute(self, siteName, pendingSlots, conn = None, transaction = False):
        self.dbi.processData(self.sql, {"site": siteName, "slots": pendingSlots},
                             conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_GetSiteInfo_

Oracle implementation of Locations.GetSiteInfo
"""

__all__ = []



from WMCore.WMBS.MySQL.Locations.GetSiteInfo import GetSiteInfo as MySQLGetSiteInfo

class GetSiteInfo(MySQLGetSiteInfo):
    """
    Same as MySQL version

    """



#!/usr/bin/env python
"""
_JobGroup_

Definition of JobGroup:
    Set of jobs running on same input file for same Workflow
    Set of jobs for a single subscription
    Required for certain job splitting Algo's (.g. event split to make complete
    lumi)
    Subscription:JobGroup == 1:N
    JobGroup:Jobs = 1:N
    JobGroup:InFile = 1:1
    JobGroup:MergedOutFile = N:1
    JobGroup at least one Lumi section

A JobGroup is a set of jobs and a Fileset that contains their output.

JobGroup knows the Subscription and passes the Workflow to Jobs in the group.

Jobs know their status (active, failed, complete) and know the files they run
on but don't know the group. They do know their subscription and corresponding
workflow. This means Jobs can update their state in the database without
talking to the group, and WMBS JobGroups can calculate status from the database
instead of the in memory objects.

The group has a status call which goes through the jobs and updates the db for
state changes and then returns the status of the group (active, failed,
complete).

WMAgent deals with groups and calls group.status periodically
"""
from __future__ import print_function

import datetime

from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.WMObject import WMObject


class JobGroup(WMObject):
    """
    JobGroups are sets of jobs running on files who's output needs to be merged
    together.
    """

    def __init__(self, subscription=None, jobs=None):
        self.jobs = []
        self.newjobs = []
        self.id = 0

        if isinstance(jobs, list):
            self.newjobs = jobs
        elif jobs is not None:
            self.newjobs = [jobs]

        self.subscription = subscription
        self.output = Fileset()
        self.last_update = datetime.datetime.now()

    def add(self, job):
        """
        _add_

        Add a Job or list of jobs to the JobGroup.
        """
        jobList = self.makelist(job)
        self.newjobs.extend(jobList)
        return

    def commit(self):
        """
        _commit_

        Move any jobs in the newjobs dict to the job dict.  Empty the newjobs
        dict.
        """
        self.jobs.extend(self.newjobs)
        self.newjobs = []

    def commitBulk(self):
        """
        Dummy method for consistency with WMBS implementation
        """
        self.commit()

    def addOutput(self, file):
        """
        _addOutput_

        Add a File to the JobGroup's output fileset.  The File is committed
        to the Fileset immediately.
        """
        self.output.addFile(file)
        self.output.commit()

    def getJobs(self, type="list"):
        """
        _getJobs_

        Retrieve all of the jobs in the JobGroup.  The output will either be
        returned as a list of Job objects (when type is "list") or a list of
        Job IDs (when type is "id").
        """
        if type == "list":
            return self.jobs
        elif type == "id":
            jobIDs = []

            for job in self.jobs:
                jobIDs.append(job["id"])

            return jobIDs
        else:
            print("Unknown type: %s" % type)

        return

    def getOutput(self, type="list"):
        """
        _getOutput_

        Retrieve all of the files that are in the JobGroup's output fileset.
        Type can be one of the following: list, set, lfn, id.
        """
        return self.output.getFiles(type=type)

    def getLength(self, obj):
        """
        This just gets a length for either dict or list objects
        """
        if isinstance(obj, (dict, list)):
            return len(obj)
        else:
            return 0

    def __len__(self):
        """
        Allows use of len() on JobGroup
        """
        return self.getLength(self.jobs) + self.getLength(self.newjobs)

        # return len(self.jobs.keys()) + len(self.newjobs.keys())



"""
_JobTypeCountByState_

Oracle implementation of Monitoring.JobTypeCountByState
"""
from __future__ import print_function, division

from WMCore.WMBS.MySQL.Monitoring.JobTypeCountByState import JobTypeCountByState as MySQLJobTypeCountByState

class JobTypeCountByState(MySQLJobTypeCountByState):
    pass



#!/usr/bin/env python
"""
_BossAirPlugin_

Base class for BossAir plugins
"""
from __future__ import division
from builtins import range

import os
import logging
from WMCore.BossAir.Plugins.BasePlugin import BasePlugin, BossAirPluginException
from datetime import datetime
from datetime import timedelta
from random import randint
import multiprocessing
import pickle

def processWorker(myinput, tmp):
    try:
        while True:
            jj, report, lcreport = myinput.get()

            if jj == 'STOP':
                return

            targetDir = jj['cache_dir']
            outfile = os.path.join(targetDir,"Report.0.pkl")

            if os.path.isfile(outfile):
                continue

            taskName = targetDir.split('/')[5]
            if jj['cache_dir'].count("Production/LogCollect") > 0:
                if lcreport is not None:
                    lcreport.task = "/" + taskName + "/Production/LogCollect"
                    with open(outfile, 'wb') as f:
                        logging.debug('Process worker is dumping the LogCollect report to ' + f.name)
                        pickle.dump(lcreport, f)
                    continue
                else:
                    msg = "Parameter lcFakeReport is mandatory if you are using logCollect jobs"
                    raise BossAirPluginException(msg)

            #ensure each lfn of each output file in the job is unique by adding the jobid
            jobid = str(jj['id'])

            if hasattr(report, 'cmsRun1') and hasattr(report.cmsRun1.output, 'output'):
                tmpname = report.cmsRun1.output.output.files.file0.lfn.split('.root')[0]
                tmpname = tmpname + jobid
                report.cmsRun1.output.output.files.file0.lfn = tmpname + '.root'

            if hasattr(report, 'logArch1') and hasattr(report.logArch1, 'output'):
                tmpname = report.logArch1.output.logArchive.files.file0.lfn.split('.tar.gz')[0]
                tmpname = tmpname + jobid
                report.logArch1.output.logArchive.files.file0.lfn = tmpname + '.root'

            #get target diretory and set task name
            report.task = "/" + taskName + "/Production"

            #pickle the report again
            with open(outfile, 'wb') as f:
                logging.debug('Process worker is dumping the report to ' + f.name)
                pickle.dump(report, f)
    except Exception as ex:
        logging.exception(ex)


class MockPlugin(BasePlugin):
    """
    Implementation of a plugin which make fake submission to the Grid
    """
    @staticmethod
    def stateMap():
        """
        For a given name, return a global state
        """

        stateDict = {'New': 'Pending',
                     'Timeout': 'Error',
                     'Submitted': 'Pending',
                     'Waiting': 'Pending',
                     'Ready': 'Pending',
                     'Scheduled': 'Pending',
                     'Running': 'Running',
                     'Done(failed)': 'Running',
                     'Done': 'Running',
                     'Aborted': 'Error',
                     'Cleared': 'Complete',
                     'Cancelled by user': 'Complete',
                     'Cancelled': 'Error'
        }

        # This call is optional but needs to for testing
        BasePlugin.verifyState(stateDict)
        return stateDict

    def __init__(self, config):

        BasePlugin.__init__(self, config)
        self.config = config

        self.pool = []
        self.createdReport = []
        self.myinput = None
        if getattr(config.BossAir, 'MockPlugin', None) == None:
            msg = "Missing required config.BossAir.MockPlugin section"
            raise BossAirPluginException( msg )

        self.nProcess = getattr(config.BossAir.MockPlugin, 'mockPluginProcesses', 4)
        self.jobRunTime = getattr(config.BossAir.MockPlugin, 'jobRunTime', 120) #default job running time is two hours
        logging.info('Job Running time set to minutes %s' % self.jobRunTime)

        self.fakeReport = getattr(config.BossAir.MockPlugin, 'fakeReport', None)
        if self.fakeReport == None:
            msg = 'config.BossAir.MockPlugin.fakeReport is a required parameter'
            raise BossAirPluginException( msg )
        elif not os.path.isfile(self.fakeReport):
            msg = 'Cannot find %s file' % self.fakeReport
            raise BossAirPluginException( msg )

        self.jobsScheduledEnd = {}

        self.states = [ 'New',
                        'Timeout',
                        'Submitted',
                        'Waiting',
                        'Ready',
                        'Scheduled',
                        'Running',
                        'Done(failed)',
                        'Done',
                        'Aborted',
                        'Cleared',
                        'Cancelled by user',
                        'Cancelled']


    def submit(self, jobs, info = None):
        """
        The submit does nothing. It simply return all jobs as successful
        """

        #All jobs are succsessfully submitted ;)
        return jobs, []



    def track(self, jobs, info = None, currentTime = None):
        """
        Label the jobs as done

        """
        runningList = []
        completeList = []
        changeList = []

        #Create the workers if they do not exists
        if self.myinput == None:
            self.myinput  = multiprocessing.Queue()
            self.start( self.myinput )

        #for each job we will need to modify the default Report (the output of each job).
        with open(self.fakeReport, "rb") as f:
            report = pickle.load(f)

        lcreport = getattr(self.config.BossAir.MockPlugin, 'lcFakeReport', None)
        if lcreport != None:
            with open(lcreport, "rb") as f:
                lcreport = pickle.load(f)

        for jj in jobs:
            if jj['id'] not in self.jobsScheduledEnd:
                self._scheduleJob(jj, currentTime)
            oldState = jj['status']
            jobEnded = datetime.now() > self.jobsScheduledEnd[jj['id']]
            if jobEnded:
                jj['globalState'] = MockPlugin.stateMap()['Done']
                jj['status'] = 'Done'
                completeList.append(jj)
            else:
                jj['globalState'] = MockPlugin.stateMap()['Running']
                jj['status'] = 'Running'
                runningList.append(jj)
            if oldState != jj['status']:
                changeList.append(jj)

            if not jj['id'] in self.createdReport:
                #Copy a fake Report in the directory using workers
                self.myinput.put((jj, report, lcreport))
                self.createdReport.append(jj['id'])

        return runningList, changeList, completeList

    def kill(self, jobs, raiseEx=False):
        """
        _kill_

        Do nothing
        """
        pass

    def complete(self, jobs):
        """
        We do not need to do anything. The report has been copied in the track
        """

        return

    def _scheduleJob(self, job, currentTime = None):
        """
        Schedule the endtime of the job
        """
        if currentTime:
            # someone told us when to start
            nowt = currentTime
        else:
            nowt = datetime.now()
        #Compute some random (between 0 and 20% of the total running time)
        randlen = randint(0, int(self.jobRunTime * 20 / 100))
        totlen = randlen + self.jobRunTime
        self.jobsScheduledEnd[job['id']] = nowt + timedelta(minutes = totlen)
        logging.debug('Scheduled end time of job %s to %s' % (job['id'], self.jobsScheduledEnd[job['id']]))



    def start( self, myinput ):
        """
        _start_

        Start the mulitp.
        """
        if len(self.pool) == 0:
            # Starting things up
            for x in range(self.nProcess):
                logging.debug("Starting process %i" %x)
                p = multiprocessing.Process(target = processWorker,
                                           args = (myinput, 0))
                p.start()
                self.pool.append(p)

    def close(self, input, result):
        """
        _close_

        Kill all connections and terminate
        """
        logging.debug("Ready to close all %i started processes " \
                        % len(self.pool) )
        for x in self.pool:
            try:
                logging.debug("Shutting down %s " % str(x))
                input.put( ('-1', 'STOP', 'control') )
            except Exception as ex:
                msg =  "Hit some exception in deletion\n"
                msg += str(ex)
                logging.error(msg)

        for proc in self.pool:
            proc.terminate()

        self.pool = []
        logging.debug('Slave stopped!')
        return


    def __del__(self):
        self.close(self.myinput, None)

    def updateSiteInformation(self, jobs, siteName, excludeSite) :
        """
        almost do nothing
        """
        return jobs



#!/usr/bin/env python
"""
_FailInput_

Oracle implementation of Jobs.FailInput
"""




from WMCore.WMBS.MySQL.Jobs.FailInput import FailInput as MySQLFailInput

class FailInput(MySQLFailInput):
    """
    _FailInput_

    The file columns names are different in Oracle, we need to take that into
    account in these queries.
    """
    fileSelect = """SELECT wmbs_jobgroup.subscription AS subid,
                           wmbs_job_assoc.fileid AS fileid FROM wmbs_job_assoc
                      INNER JOIN wmbs_job ON
                        wmbs_job_assoc.job = wmbs_job.id
                      INNER JOIN wmbs_jobgroup ON
                        wmbs_job.jobgroup = wmbs_jobgroup.id
                      LEFT OUTER JOIN wmbs_sub_files_failed ON
                        wmbs_jobgroup.subscription = wmbs_sub_files_failed.subscription AND
                        wmbs_job_assoc.fileid = wmbs_sub_files_failed.fileid
                    WHERE wmbs_job.id = :jobid AND wmbs_sub_files_failed.fileid IS Null"""

    acquiredDelete = """DELETE FROM wmbs_sub_files_acquired
                        WHERE subscription = :subid AND fileid = :fileid"""

    completeDelete = """DELETE FROM wmbs_sub_files_complete
                      WHERE subscription = :subid AND fileid = :fileid"""

    sql = """INSERT INTO wmbs_sub_files_failed (fileid, subscription)
               SELECT :fileid, :subid FROM dual
               WHERE NOT EXISTS (SELECT fileid FROM wmbs_sub_files_failed wff2
                                   WHERE wff2.fileid = :fileid
                                   AND wff2.subscription = :subid)"""



from future.utils import viewitems

import cherrypy
from os import listdir

from cherrypy import expose, tools

from WMCore.WebTools.FrontEndAuth import get_user_info
from WMCore.WebTools.Page import TemplatedPage


class SecureDocumentation(TemplatedPage):
    """
    The documentation for the framework
    """

    @expose
    @tools.secmodv2()
    def index(self):
        templates = listdir(self.templatedir)
        user = get_user_info()
        index = "<h1>Secure Documentation</h1>"
        index += "<div>\n"
        index += "Hello <b>%s</b>!\n" % user['name']
        index += "<ul>\n"
        index += "<li>Your <b>DN</b>: %s</li>\n" % user['dn']
        index += "<li>Your <b>Login</b>: %s</li>\n" % user['login']
        index += "<li>Your <b>Authn Method</b>: %s</li>\n" % user['method']
        index += "<li>Your roles:\n"
        index += "<ul>\n"
        for k, v in viewitems(user['roles']):
            for t in ['group', 'site']:
                for n in v[t]:
                    index += "<li><b>%s</b>: %s=%s</li>\n" % (k, t, n)
        index += "</ul></li></ul>\n"
        index += "<a href='CernDocs'>CERN docs</a>\n"
        index += "<a href='UerjDocs'>UERJ docs</a>\n"
        index += "</div>\n"

        index += "<ol>"
        for t in templates:
            if '.tmpl' in t:
                index = "%s\n<li><a href='%s'>%s</a></li>" % (index,
                                                              t.replace('.tmpl', ''),
                                                              t.replace('.tmpl', ''))
        index = "%s\n<li><a href='https://twiki.cern.ch/twiki/bin/view/CMS/DMWebtools'>twiki</a>" % (index)
        index = "%s\n</ol>" % (index)

        return index

    @expose
    @tools.secmodv2(role='admin')
    def AdminDocs(self):
        return "This is a secret documentation for admins."

    @expose
    @tools.secmodv2(site='t1-ch-cern')
    def CernDocs(self):
        return "This is a secret documentation about the T1 CERN site"

    @expose
    @tools.secmodv2(role='dev', group='dmwm')
    def DevDmwmDocs(self):
        return "This is a secret documentation for DMWM developers"

    @expose
    def default(self, *args, **kwargs):
        if len(args) > 0:
            return self.templatepage(args[0])
        return self.index()



#!/usr/bin/env python
"""
_WorkerThreadManager_

A class used to manage regularly running worker threads.
"""
from __future__ import print_function

from builtins import object
import logging
import threading
import time

from WMCore.Agent.HeartbeatAPI import HeartbeatAPI
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread

# keep track of a unique WTM number
wtmcount = 0


class WorkerThreadManager(object):
    """
    Manages regular worker slave threads
    """

    def __init__(self, component):
        """
        Set up the events used to pause, resume and terminate worker threads
        """
        global wtmcount
        self.component = component
        self.terminateSlaves = threading.Event()
        self.pauseSlaves = threading.Event()
        self.resumeSlaves = threading.Event()
        self.lock = threading.Lock()
        self.lock.acquire()
        self.activeThreadCount = 0
        self.wtmnumber = wtmcount
        wtmcount = wtmcount + 1
        self.slavecounter = 0
        self.slavelist = []
        self.lock.release()
        logging.info("Started")
        return

    def slaveTerminateCallback(self, slaveid):
        """
        Callback function invoked by terminated slave threads
        """
        self.lock.acquire()
        try:
            self.slavelist.remove("threadmanager-slave%s" % slaveid)
        except:
            pass
        else:
            self.activeThreadCount -= 1
        self.lock.release()

    def prepareWorker(self, worker, idleTime, heartbeatTimeout):
        """
        Prepares a worker thread before running
        """
        # Work timing
        worker.idleTime = idleTime
        worker.component = self.component
        self.lock.acquire()
        self.slavecounter += 1
        worker.slaveid = "%s-%s" % (self.wtmnumber, self.slavecounter)
        self.lock.release()

        # Thread synchronisation
        worker.notifyTerminate = self.terminateSlaves
        worker.terminateCallback = self.slaveTerminateCallback
        worker.notifyPause = self.pauseSlaves
        worker.notifyResume = self.resumeSlaves
        if hasattr(self.component.config, "Agent"):
            if getattr(self.component.config.Agent, "useHeartbeat", True):
                worker.heartbeatAPI = HeartbeatAPI(self.component.config.Agent.componentName,
                                                   idleTime, heartbeatTimeout)

    def addWorker(self, worker, idleTime=60, hbTimeout=None, parameters=None):
        """
        Adds a worker object and sets it running. Worker thread will sleep for
        idleTime seconds between runs. Parameters, if present, are passed into
        the worker thread's setup, algorithm and terminate methods
        """
        # Check type of worker
        if not isinstance(worker, BaseWorkerThread):
            msg = "Attempting to add worker that does not inherit from "
            msg += "BaseWorkerThread"
            logging.critical(msg)
            return

        # Prepare the new worker thread
        self.prepareWorker(worker, idleTime, hbTimeout)
        workerThread = threading.Thread(target=worker, args=(parameters,))
        msg = "Created worker thread %s" % str(worker)
        logging.info(msg)

        # Increase the active thread count - note this must be done before
        # starting the thread so the callback can decrease back in case of
        # startup failure
        self.lock.acquire()
        self.activeThreadCount += 1
        workerThread.name = "threadmanager-slave%s" % worker.slaveid
        self.slavelist.append(workerThread.name)
        self.lock.release()

        # Actually start the thread
        workerThread.start()

    def terminateWorkers(self):
        """
        Terminates all threads
        """
        # Don't change order without looking at BaseWorkerThread!
        # Notify all threads to terminate
        self.terminateSlaves.set()
        self.pauseSlaves.clear()
        self.resumeSlaves.set()

        # Wait for all threads to finished
        finished = False
        while not finished:
            self.lock.acquire()
            logging.info("Waiting for %s worker threads to terminate", self.activeThreadCount)
            logging.debug("\n slavelist is %s", self.slavelist)
            logging.debug("\n threadlist is %s", threading.enumerate())

            if self.activeThreadCount == 0:
                finished = True
            else:
                # check to make sure we aren't waiting on dead threads
                threadlist = threading.enumerate()
                # we want to look for all the slavenames
                #  that correspond to nonexistant or dead threads
                # also, I realize this is O(N^2), but my brain hurts too hard
                #  to do it nicer

                # this (should be) race-proof. my thoughts:
                #  this is in a lock, the only other places that modify
                #  activeThreadCount are within locks
                # everywhere else that tries to modify active thread count
                # also does it with a try-except-else around removing from
                # slavelist. Slavelist becomes our synchronization
                for slavename in self.slavelist:
                    found = False
                    for threadobj in threadlist:
                        if hasattr(threadobj, 'name') and (slavename == threadobj.name) and (threadobj.is_alive()):
                            found = True
                    if found is False:
                        # the slave we wanted wasn't running
                        try:
                            self.slavelist.remove(slavename)
                        except Exception as ex:
                            print("couldn't remove thread.. %s " % ex)
                        else:
                            self.activeThreadCount -= 1
            self.lock.release()
            time.sleep(5)
        logging.info("All worker threads terminated")

    def pauseWorkers(self):
        """
        Pauses all running threads
        """
        # Don't change order without looking at BaseWorkerThread!
        self.resumeSlaves.clear()
        self.pauseSlaves.set()
        logging.info("All worker threads paused")

    def resumeWorkers(self):
        """
        Resumes all running threads
        """
        # Don't change order without looking at BaseWorkerThread!
        self.pauseSlaves.clear()
        self.resumeSlaves.set()
        logging.info("All worker threads resumed")



#!/usr/bin/env python
"""
_LoadFromSpecOwner_

Oracle implementation of Workflow.LoadFromSpecOwner

"""
__all__ = []



from WMCore.WMBS.MySQL.Workflow.LoadFromSpecOwner import LoadFromSpecOwner \
     as LoadWorkflowMySQL

class LoadFromSpecOwner(LoadWorkflowMySQL):
    sql = LoadWorkflowMySQL.sql



#!/usr/bin/env python
"""
_CheckForDelete_

MySQL implementation of DeleteCheck

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class CheckForDelete(DBFormatter):
    sql = """SELECT wf.id FROM wmbs_fileset wf
               WHERE wf.id = :fileset
               AND NOT EXISTS (SELECT id FROM wmbs_subscription ws
                               WHERE ws.fileset = :fileset
                               AND ws.id != :subscription)
                               """

    def execute(self, fileids = None, subid = None, conn = None, transaction = False):
        """
        _execute_

        Given a list of fileset IDs, figure out which ones can be deleted
        """
        binds = []
        for id in fileids:
            binds.append({'fileset': id, 'subscription': subid})

        result = self.dbi.processData(self.sql, binds,
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_LoadForMonitoring_

MySQL implementation for loading a job by scheduler status
"""


from WMCore.Database.DBFormatter import DBFormatter

class LoadForMonitoring(DBFormatter):
    """
    _LoadForMonitoring_

    Load all jobs with a certain scheduler status including
    all the joined information.
    """


    sql = """SELECT rj.wmbs_id AS jobid, rj.grid_id AS gridid, rj.bulk_id AS bulkid,
               st.name AS status, rj.retry_count as retry_count, rj.id AS id,
               rj.status_time as status_time, wl.plugin AS plugin, wu.cert_dn AS owner
               FROM bl_runjob rj
               INNER JOIN bl_status st ON rj.sched_status = st.id
               LEFT OUTER JOIN wmbs_users wu ON wu.id = rj.user_id
               INNER JOIN wmbs_job wj ON wj.id = rj.wmbs_id
               LEFT OUTER JOIN wmbs_location wl ON wl.id = wj.location
               WHERE rj.status = :complete
    """



    def execute(self, complete = '1', conn = None, transaction = False):
        """
        _execute_

        Load all jobs either running or not (running by default)
        """

        binds = {'complete': complete}


        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_ListRunsWorkflow_

For a given workflow, list the processed runs according to DBSBUFFER tables
"""

from future.utils import viewvalues

import threading

from WMCore.Database.DBFormatter import DBFormatter

class ListRunsWorkflow(DBFormatter):

    sql = """SELECT DISTINCT dbsbuffer_file_runlumi_map.run FROM
                dbsbuffer_file_runlumi_map INNER JOIN dbsbuffer_file
                ON dbsbuffer_file_runlumi_map.filename = dbsbuffer_file.id
                INNER JOIN dbsbuffer_workflow
                ON dbsbuffer_file.workflow = dbsbuffer_workflow.id
                WHERE dbsbuffer_workflow.name = :workflow"""

    def execute(self, workflow = None, conn = None, transaction = False):
        """
        _execute_

        Changed to expect a DBSBlock object
        """
        bindVars = []

        bindVars.append({"workflow": workflow})

        results = self.dbi.processData(self.sql, bindVars, conn = conn,
                             transaction = transaction)
        runs = []

        for result in results[0].fetchall():
            runs.append(next(iter(viewvalues(result))))

        return runs



#!/usr/bin/env python
"""
_SetLocationByLFN_

Oracle implementation of Files.SetLocationByLFN
"""




from WMCore.WMBS.MySQL.Files.SetLocationByLFN import SetLocationByLFN as MySQLSetLocationByLFN

class SetLocationByLFN(MySQLSetLocationByLFN):
    pass



#!/usr/bin/env python
"""
_NewAlgo_

MySQL implementation of DBSBuffer.NewAlgo
"""




from WMCore.Database.DBFormatter import DBFormatter

class NewAlgo(DBFormatter):

    sql = """INSERT IGNORE INTO dbsbuffer_algo
             (app_name, app_ver, app_fam, pset_hash, config_content, in_dbs)
             VALUES (:app_name, :app_ver, :app_fam, :pset_hash, :config_content, 0)
             """

    def execute(self, appName, appVer, appFam, psetHash = None,
                configContent = None, conn = None, transaction = False):
        """
        _execute_

        Add a new algo into the database
        """
        binds = {"app_name": appName, "app_ver": appVer, "app_fam": appFam,
                 "pset_hash": psetHash, "config_content": configContent}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python

"""
_Step.Executor.CMSSW_

Implementation of an Executor for a CMSSW step.
"""

import logging
import os
import socket
import subprocess
import sys
import time
from resource import getrusage, RUSAGE_SELF, RUSAGE_CHILDREN

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytesConditional
from WMCore.FwkJobReport.Report import addAttributesToFile
from WMCore.WMExceptions import WM_JOB_ERROR_CODES
from WMCore.WMRuntime.Tools.Scram import Scram
from WMCore.WMRuntime.Tools.Scram import getSingleScramArch
from WMCore.WMSpec.Steps.Executor import Executor
from WMCore.WMSpec.Steps.WMExecutionFailure import WMExecutionFailure
from WMCore.WMSpec.WMStep import WMStepHelper


def analysisFileLFN(fileName, lfnBase, job):
    """
    Construct an LFN for a user file
    """

    base = os.path.split(fileName)[1]
    root, ext = os.path.splitext(base)

    newBase = '{base}_{count:04d}{ext}'.format(base=root, ext=ext, count=job['counter'])
    lfn = os.path.join(lfnBase, job['workflow'], 'output', newBase)

    return lfn


class CMSSW(Executor):
    """
    _CMSWW_

    Execute a CMSSW Step

    """

    def __init__(self):
        super(CMSSW, self).__init__()
        self.failedPreviousStep = None

    def _setStatus(self, returnCode, returnMessage):
        """Chirp Job/step exit code through HTCondor"""
        # we will set Chirp_WMCore_cmsRun_ExitCode codes only if previous step was clean
        if not self.failedPreviousStep:
            self.setCondorChirpAttrDelayed('Chirp_WMCore_cmsRun_ExitCode', returnCode)
            logging.info("Step %s: Chirp_WMCore_cmsRun_ExitCode %s", self.stepName, returnCode)
            if returnMessage and returnCode != 0:
                self.setCondorChirpAttrDelayed('Chirp_WMCore_cmsRun_Exception_Message', returnMessage, compress=True)
        self.setCondorChirpAttrDelayed('Chirp_WMCore_%s_ExitCode' % self.stepName, returnCode)
        logging.info("Step %s: Chirp_WMCore_%s_ExitCode %s", self.stepName, self.stepName, returnCode)
        if returnMessage and returnCode != 0:
            self.setCondorChirpAttrDelayed('Chirp_WMCore_%s_Exception_Message' % self.stepName, returnMessage, compress=True)

        self.step.execution.exitStatus = returnCode

    def pre(self, emulator=None):
        """
        _pre_

        Pre execution checks

        """
        if emulator is not None:
            return emulator.emulatePre(self.step)

        logging.info("Steps.Executors.%s.pre called", self.__class__.__name__)

        if hasattr(self.step.application.configuration, 'configCacheUrl'):
            # means we have a configuration & tweak in the sandbox
            psetFile = self.step.application.command.configuration
            psetTweak = self.step.application.command.psetTweak
            self.stepSpace.getFromSandbox(psetFile)

            if psetTweak:
                self.stepSpace.getFromSandbox(psetTweak)

        if hasattr(self.step, "pileup"):
            self.stepSpace.getFromSandbox("pileupconf.json")

        # add in ths scram env PSet manip script whatever happens
        self.step.runtime.preScripts.append("SetupCMSSWPset")
        return None

    def execute(self, emulator=None):
        """
        _execute_

        """
        if emulator is not None:
            return emulator.emulate(self.step, self.job)

        logging.info("Steps.Executors.%s.execute called", self.__class__.__name__)

        stepModule = "WMTaskSpace.%s" % self.stepName

        overrides = {}
        if hasattr(self.step, 'override'):
            overrides = self.step.override.dictionary_()
        self.failedPreviousStep = overrides.get('previousCmsRunFailure', False)

        if self.failedPreviousStep:
            # the previous cmsRun step within this task failed
            # don't bother executing anything else then
            msg = WM_JOB_ERROR_CODES[99108]
            logging.critical(msg)
            self._setStatus(99108, msg)
            raise WMExecutionFailure(99108, "CmsRunFailure", msg)

        # write the wrapper script to a temporary location
        # I don't pass it directly through os.system because I don't
        # trust that there won't be shell-escape shenanigans with
        # arbitrary input files
        scramSetup = self.step.application.setup.softwareEnvironment
        scramCommand = self.step.application.setup.scramCommand
        scramProject = self.step.application.setup.scramProject
        scramArch = self.step.application.setup.scramArch
        cmsswVersion = self.step.application.setup.cmsswVersion
        jobReportXML = self.step.output.jobReport
        cmsswCommand = self.step.application.command.executable
        cmsswConfig = self.step.application.command.configuration
        cmsswArguments = self.step.application.command.arguments

        scramArch = getSingleScramArch(scramArch)

        try:
            multicoreSettings = self.step.application.multicore
            logging.info("CMSSW configured for %s cores and %s event streams",
                         multicoreSettings.numberOfCores, multicoreSettings.eventStreams)
        except AttributeError:
            logging.info("No value set for multicore numberOfCores or eventStreams")

        try:
            gpuSettings = self.step.application.gpu
            logging.info("CMSSW configured for GPU required: %s, with these settings: %s",
                         gpuSettings.gpuRequired, gpuSettings.gpuRequirements)
        except AttributeError:
            logging.info("No value set for GPU gpuRequired and/or gpuRequirements")

        logging.info("Executing CMSSW step")

        #
        # set any global environment variables
        #
        try:
            os.environ['FRONTIER_ID'] = 'wmagent_%s' % (self.report.data.workload)
        except Exception as ex:
            logging.error('Have critical error in setting FRONTIER_ID: %s', str(ex))
            logging.error('Continuing, as this is not a critical function yet.')

        #
        # scram bootstrap
        #
        scram = Scram(
            command=scramCommand,
            version=cmsswVersion,
            initialise=self.step.application.setup.softwareEnvironment,
            directory=self.step.builder.workingDir,
            architecture=scramArch,
        )

        logging.info("Runing SCRAM")
        try:
            projectOutcome = scram.project()
        except Exception as ex:
            msg = WM_JOB_ERROR_CODES[50513]
            msg += "\nDetails: %s" % str(ex)
            logging.critical(msg)
            raise WMExecutionFailure(50513, "ScramSetupFailure", msg)
        if projectOutcome > 0:
            msg = WM_JOB_ERROR_CODES[50513]
            msg += "\nDetails: %s" % str(scram.diagnostic())
            logging.critical(msg)
            raise WMExecutionFailure(50513, "ScramSetupFailure", msg)

        runtimeOutcome = scram.runtime()
        if runtimeOutcome > 0:
            msg = WM_JOB_ERROR_CODES[50513]
            msg += "\nDetails: %s" % str(scram.diagnostic())
            logging.critical(msg)
            raise WMExecutionFailure(50513, "ScramSetupFailure", msg)

        #
        # pre scripts
        #
        logging.info("Running PRE scripts")
        for script in self.step.runtime.preScripts:
            # TODO: Exception handling and error handling & logging
            scriptProcess = subprocess.Popen(
                ["/bin/bash"], shell=True, cwd=self.step.builder.workingDir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )
            # BADPYTHON
            invokeCommand = "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH\n"
            invokeCommand += "{} -m WMCore.WMRuntime.ScriptInvoke {} {} \n".format(sys.executable,
                                                                                   stepModule,
                                                                                   script)
            logging.info("    Invoking command:\n%s", invokeCommand)
            scriptProcess.stdin.write(encodeUnicodeToBytesConditional(invokeCommand, condition=PY3))
            stdout, stderr = scriptProcess.communicate()
            retCode = scriptProcess.returncode
            if retCode > 0:
                msg = "Error running command\n%s\n" % invokeCommand
                msg += "%s\n %s\n %s\n" % (retCode, stdout, stderr)
                logging.critical("Error running command")
                logging.critical(msg)
                raise WMExecutionFailure(50513, "PreScriptFailure", msg)

        #
        # pre scripts with scram
        #
        logging.info("RUNNING SCRAM SCRIPTS")
        for script in self.step.runtime.scramPreScripts:
            # invoke scripts with scram()
            runtimeDir = getattr(self.step.runtime, 'scramPreDir', None)
            invokeCommand = self.step.runtime.invokeCommand if hasattr(self.step.runtime, 'invokeCommand') else \
                "%s -m WMCore.WMRuntime.ScriptInvoke %s" % (sys.executable, stepModule)
            invokeCommand += " %s \n" % script
            retCode = scram(invokeCommand, runtimeDir=runtimeDir)
            if retCode > 0:
                msg = "Error running command\n%s\n" % invokeCommand
                msg += "%s\n " % retCode
                msg += scram.diagnostic()
                logging.critical(msg)
                raise WMExecutionFailure(50513, "PreScriptScramFailure", msg)

        configPath = "%s/%s-main.sh" % (self.step.builder.workingDir, self.stepName)
        with open(configPath, 'w') as handle:
            handle.write(CONFIG_BLOB)

        # spawn this new process
        # the script looks for:
        # <SCRAM_COMMAND> <SCRAM_PROJECT> <CMSSW_VERSION> <JOB_REPORT> <EXECUTABLE> <CONFIG>
        args = ['/bin/bash',
                configPath,
                scramSetup,
                scramArch,
                scramCommand,
                scramProject,
                cmsswVersion,
                jobReportXML,
                cmsswCommand,
                cmsswConfig,
                cmsswArguments]
        logging.info("Executing CMSSW. args: %s", args)

        # possibly needed environment overrides for CMSSW call go here
        envOverride = {}
        # Do not pass WM PYTHONPATH to CMSSW environment
        pythonPath = os.environ.get('PYTHONPATH', '')
        envOverride['PYTHONPATH'] = ""
        # work around problem with GSI authentication plugin and EOS at CERN
        if socket.getfqdn().endswith("cern.ch"):
            envOverride['XRD_LOADBALANCERTTL'] = "86400"
        # some libraries linked with CMSSW need HOME in the environment
        if 'HOME' not in os.environ:
            envOverride['HOME'] = os.environ.get('PWD', "/")

        os.environ.update(envOverride)

        # take snapshot of user and sys time before we call subprocess
        startTime = time.time()
        userTime0 = getrusage(RUSAGE_CHILDREN).ru_utime
        sysTime0 = getrusage(RUSAGE_CHILDREN).ru_stime
        returnCode = 1  # by default we assume process may fail
        with open(self.step.output.stdout, 'w') as stdoutHandle, open(self.step.output.stderr, 'w') as stderrHandle:
            proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            childPid, exitStatus, resource = os.wait4(proc.pid, os.P_PID)
            stdout, stderr = proc.communicate()
            stdoutHandle.write(str(stdout))
            stderrHandle.write(str(stderr))
            returnCode = proc.returncode

        # calculate user and sys time of subprocess by substructing relevant
        # parts take before we spawn the subprocess
        userTime = getrusage(RUSAGE_CHILDREN).ru_utime - userTime0
        sysTime = getrusage(RUSAGE_CHILDREN).ru_stime - sysTime0
        endTime = time.time()
        returnMessage = None

        # Return PYTHONPATH to its original value, as this
        # is needed for stepChain workflows, so other prescripts
        # are able to find WMCore modules
        envOverride['PYTHONPATH'] = pythonPath
        os.environ.update(envOverride)

        if returnCode != 0:
            argsDump = {'arguments': args}
            msg = "Error running cmsRun\n%s\n" % argsDump
            try:
                self.report.parse(jobReportXML, stepName=self.stepName)
                (returnCode, returnMessage) = self.report.getStepExitCodeAndMessage(stepName=self.stepName)
                msg += "CMSSW Return code: %s\n" % returnCode
            except Exception as ex:
                # If report parsing fails, report linux exit code
                msg += "Linux Return code: %s\n" % returnCode
            finally:
                logging.critical(msg)
                logging.critical("Error message: %s", returnMessage)
                self._setStatus(returnCode, returnMessage)
                raise WMExecutionFailure(returnCode, "CmsRunFailure", msg)
        else:
            self._setStatus(returnCode, returnMessage)

        try:
            # update FJR report info with subprocess info
            self.report.updateSubprocessInfo(sysTime, userTime, startTime, endTime)

            # parse job report XML
            self.report.parse(jobReportXML, stepName=self.stepName)
        except Exception as ex:
            msg = WM_JOB_ERROR_CODES[50115]
            msg += "\nDetails: %s" % str(ex)
            raise WMExecutionFailure(50115, "BadJobReportXML", msg)

        stepHelper = WMStepHelper(self.step)
        typeHelper = stepHelper.getTypeHelper()

        acquisitionEra = typeHelper.getAcqEra() or self.task.getAcquisitionEra()
        processingVer = typeHelper.getProcVer() or self.task.getProcessingVersion()
        processingStr = typeHelper.getProcStr() or self.task.getProcessingString()
        prepID = typeHelper.getPrepId() or self.task.getPrepID()
        globalTag = typeHelper.getGlobalTag()
        validStatus = self.workload.getValidStatus()
        inputPath = self.task.getInputDatasetPath()
        campaign = self.workload.getCampaign()
        cacheUrl, cacheDB, configID = stepHelper.getConfigInfo()

        self.report.setValidStatus(validStatus=validStatus)
        self.report.setGlobalTag(globalTag=globalTag)
        self.report.setCampaign(campaign)
        self.report.setPrepID(prepID)
        self.report.setInputDataset(inputPath=inputPath)
        self.report.setAcquisitionProcessing(acquisitionEra=acquisitionEra,
                                             processingVer=processingVer,
                                             processingStr=processingStr)
        self.report.setConfigURL(configURL="%s;;%s;;%s" % (cacheUrl,
                                                           cacheDB,
                                                           configID))

        # Attach info to files
        self.report.addInfoToOutputFilesForStep(stepName=self.stepName, step=self.step)

        self.report.checkForOutputFiles(stepName=self.stepName)
        self.report.checkForAdlerChecksum(stepName=self.stepName)
        self.report.checkForRunLumiInformation(stepName=self.stepName)

        if self.step.output.keep != True:
            self.report.killOutput()
        else:
            # Check that we only keep the desired output
            for module in stepHelper.getIgnoredOutputModules():
                self.report.deleteOutputModuleForStep(stepName=self.stepName, moduleName=module)

        # Add stageout LFN to existing TFileService files
        reportAnalysisFiles = self.report.getAnalysisFilesFromStep(self.stepName)
        for reportAnalysisFile in reportAnalysisFiles:
            newLFN = analysisFileLFN(reportAnalysisFile.fileName, self.step.user.lfnBase, self.job)
            addAttributesToFile(reportAnalysisFile, pfn=reportAnalysisFile.fileName, lfn=newLFN, validate=False)

        # Add analysis file entries for additional files listed in workflow
        for fileName in stepHelper.listAnalysisFiles():
            analysisFile = stepHelper.getAnalysisFile(fileName)
            if os.path.isfile(analysisFile.fileName):
                newLFN = analysisFileLFN(analysisFile.fileName, analysisFile.lfnBase, self.job)
                self.report.addAnalysisFile(analysisFile.fileName, lfn=newLFN, Source='UserDefined',
                                            pfn=os.path.join(os.getcwd(), analysisFile.fileName), validate=False)

        return

    def post(self, emulator=None):
        """
        _post_

        Post execution checkpointing

        """
        if emulator is not None:
            return emulator.emulatePost(self.step)

        logging.info("Steps.Executors.%s.post called", self.__class__.__name__)

        if self.report.getStepErrors(self.stepName) != {}:
            # Then we had errors
            # Go directly to spot specified in WMStep
            return self.errorDestination

        return None


CONFIG_BLOB = """#!/bin/bash

# Check to make sure the argument count is correct
REQUIRED_ARGUMENT_COUNT=5
if [ $# -lt $REQUIRED_ARGUMENT_COUNT ]
then
    echo "Usage: `basename $0` <SCRAM_SETUP>  <SCRAM_ARCH> <SCRAM_COMMAND> <SCRAM_PROJECT> <CMSSW_VERSION>\
                 <JOB_REPORT> <EXECUTABLE> <CONFIG> [Arguments for cmsRun]"
    exit 70
fi

# Extract the required arguments out, leaving an unknown number of
#  cmsRun arguments
SCRAM_SETUP=$1
SCRAM_ARCHIT=$2
SCRAM_COMMAND=$3
SCRAM_PROJECT=$4
CMSSW_VERSION=$5
JOB_REPORT=$6
EXECUTABLE=$7
CONFIGURATION=$8
shift;shift;shift;shift;shift;
shift;shift;shift;
echo "Setting up Frontier log level"
export FRONTIER_LOG_LEVEL=warning

echo "Beginning CMSSW wrapper script"
echo "$SCRAM_SETUP $SCRAM_ARCHIT $SCRAM_COMMAND $SCRAM_PROJECT"

echo "Performing SCRAM setup..."
$SCRAM_SETUP
echo "Completed SCRAM setup"

export SCRAM_ARCH=$SCRAM_ARCHIT

echo "Retrieving SCRAM project..."
# do the actual executing
$SCRAM_COMMAND project $SCRAM_PROJECT $CMSSW_VERSION
EXIT_STATUS=$?
if [ $EXIT_STATUS -ne 0 ]; then echo "Scram failed with exit code: $EXIT_STATUS"; exit 71; fi
cd $CMSSW_VERSION
EXIT_STATUS=$?
if [ $EXIT_STATUS -ne 0 ]; then echo "***\nCouldn't chdir: $EXIT_STATUS\n"; exit 72; fi

eval `$SCRAM_COMMAND runtime -sh`
EXIT_STATUS=$?
if [ $EXIT_STATUS -ne 0 ]; then echo "***\nCouldn't get scram runtime: $EXIT_STATUS\n*"; exit 73; fi
echo "Completed SCRAM project"
cd ..
echo "Executing CMSSW"
echo "$EXECUTABLE  -j $JOB_REPORT $CONFIGURATION"
$EXECUTABLE  -j $JOB_REPORT $CONFIGURATION 2>&1 &
PROCID=$!
echo $PROCID > process.id
wait $PROCID
EXIT_STATUS=$?
echo "Complete"
echo "process id is $PROCID status is $EXIT_STATUS"
exit $EXIT_STATUS

"""



#!/usr/bin/env python
"""
_DeleteFileset_

MySQL implementation of DeleteFileset

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Delete(DBFormatter):
    sql = "delete from wmbs_fileset where name = :fileset"

    def getBinds(self, name = None):
        return self.dbi.buildbinds(self.dbi.makelist(name), 'fileset')

    def execute(self, name = None, conn = None, transaction = False):
        self.dbi.processData(self.sql, self.getBinds(name),
                         conn = conn, transaction = transaction)
        return True #or raise



#!/usr/bin/env python
"""
_GFAL2Impl_
Implementation of StageOutImpl interface for gfal-copy
"""
import argparse
import logging
import os

from WMCore.Storage.Execute import runCommandWithOutput as runCommand
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutImplV2 import StageOutImplV2

_CheckExitCodeOption = True


class GFAL2Impl(StageOutImplV2):
    """
    _GFAL2Impl_
    Implement interface for GFAL2 commands (gfal-copy, gfal-rm)
    """

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum):
        """
            performs a transfer. stageOut tells you which way to go. returns the new pfn
            or raises on failure. StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin
        """
        localFileName = fromPfn
        if stageOut:
            fromPfn2 = self.prependFileProtocol(fromPfn)
            toPfn2 = toPfn
            localFileName = fromPfn
            remoteFileName = toPfn
        else:
            fromPfn2 = fromPfn
            toPfn2 = self.prependFileProtocol(toPfn)
            localFileName = toPfn
            remoteFileName = fromPfn
            localDir = os.path.dirname(localFileName)

        if not options:
            options = ''

        # parse the options
        parser = argparse.ArgumentParser()
        parser.add_argument('--nochecksum', action='store_true')
        args, unknown = parser.parse_known_args(options.split())

        baseTransferCommand = "env -i X509_USER_PROXY=$X509_USER_PROXY gfal-copy -t 2400 -T 2400 -p "
        if args.nochecksum:
            transferCommand = "%s -vvv %s %s %s " % (baseTransferCommand, ' '.join(unknown), fromPfn2, toPfn2)
        else:
            transferCommand = "%s -K adler32 -vvv %s %s %s " % (baseTransferCommand, ' '.join(unknown),
                                                                fromPfn2, toPfn2)

        logging.info("Staging out with gfal-copy")
        logging.info("  commandline: %s", transferCommand)
        commandExec = self.runCommandFailOnNonZero(transferCommand)

        if commandExec[0] != 0:
            try:
                logging.error("Transfer failed, deleting partial file")
                self.doDelete(toPfn, None, None, None, None)
            except Exception:
                pass
            raise StageOutFailure("File transfer failed")

        return toPfn

    def doDelete(self, pfn, pnn, command, options, protocol):
        """
            deletes a file, raises on error
            StageOutError (and inherited exceptions) are for expected errors
            such as temporary connection failures. Anything else will be handled as an unexpected
            error and skip retrying with this plugin
        """
        if os.path.isfile(pfn):
            runCommand("/bin/rm -f %s" % pfn)
        elif pfn.startswith("file:"):
            runCommand("env -i X509_USER_PROXY=$X509_USER_PROXY gfal-rm -vvv %s" % pfn)
        else:
            runCommand(StageOutImpl.createRemoveFileCommand(self, pfn))



#!/usr/bin/env python
"""
_LoadFromIDWithWorkflow_

Oracle implementation of Jobs.LoadFromIDWithWorkflow.
"""

from WMCore.WMBS.MySQL.Jobs.LoadFromIDWithWorkflow import LoadFromIDWithWorkflow as MySQLLoadFromIDWithWorkflow

class LoadFromIDWithWorkflow(MySQLLoadFromIDWithWorkflow):
    """
    Load jobs by ID but include the workflow name

    """

    pass



#!/usr/bin/env python
"""
_LoadFilesByBlock_

Oracle implementation of DBS3Buffer.LoadFilesByBlock
"""

from WMComponent.DBS3Buffer.MySQL.LoadFilesByBlock import LoadFilesByBlock as MySQLLoadFilesByBlock

class LoadFilesByBlock(MySQLLoadFilesByBlock):
    pass



#!/usr/bin/env python
"""
_ThreadPool_

Thread based work queue that allows a set of slaves to be added to a queue
to process information in the queue in parallel and return the output
of each slave in an output queue.

Written by Evan Jones
http://evanjones.ca/software/python-workqueue.html

Worker queues for Python (also called thread pools).

This class queues work and distributes it over a set of slave objects. The
slaves are Callable instances that get called in their own thread. I use
it with XML-RPC server instances to distribute work over a cluster of
machines.

The enqueue() function is used to queue function calls. The dequeue() function
returns the results of actually calling the function. This permits the
execution of the function to happen in the background in another thread. The
order that results are returned by dequeue() is not the same as the order work
is placed into the queue with enqueue().

This class uses "keys" to track specific work items. When enqueuing a work
item, a key is also required. The key does not affect the processing at all.
It is simply returned when dequeue() returns the result of calling the
function. This helps the caller to determine which function

"""

from builtins import object
from future import standard_library
standard_library.install_aliases()

import threading

class SerializedThreadPool( object ):
    """
    Implements the ThreadPool interface in a serialized fashion. This
    is very useful for debugging since there is no parallelism. All the
    work is performed when calling dequeue.
    """

    __slots__ = ( 'function', 'queue' )
    def __init__( self, slaveFunction ):
        """
        Creates a new SerializedThreadPool that will call the
        specified function.
        """
        self.function = slaveFunction
        if hasattr( slaveFunction, '__len__' ):
            assert( len( slaveFunction ) == 1 )
            self.function = slaveFunction[0]

        self.queue = []

    def enqueue( self, key, *parameters ):
        """
        Adds a new work item to the queue.
        This does not actually execute the function.
        """
        self.queue.append( (key, parameters) )

    def dequeue( self ):
        """
        Returns a (key, result) pair from a completed work item,
        or None if there are no more items. This will actually
        cause the function to be executed.
        """

        if len( self.queue ) == 0:
            return None

        # Take from the tail (much faster)
        key, parameters = self.queue[-1]
        del self.queue[-1]

        result = self.function( *parameters )
        return key, result

    def __iter__( self ):
        """
        Returns an iterator over the (key, result) pairs
        returned by dequeue(). The iterator stops when there
        are no more results in the queue.
        """
        result = self.dequeue()
        while result != None:
            yield result
            result = self.dequeue()

class ThreadPool( object ):
    """
    _ThreadPool_

    Distributes work over a set of slaves. Each slave is a
    callable object that will be called from a seperate thread.
    """
    def __init__( self, slaves ):
        self.lock = threading.Condition()

        # Server objects that are currently unused
        self.slaveQueue = list( slaves )
        # Pending work
        self.callQueue = []
        # Pending results
        self.resultsQueue = []
        # Number of threads currently active
        self.activeCount = 0

    def enqueue( self, key, *parameters ):
        """
        _enqueue_

        Add a new work item to the queue.
        This may result in threads being spawned.

        """

        work = (key, parameters)

        self.lock.acquire()

        # enqueue the work item
        self.callQueue.append(work)

        if len( self.slaveQueue ):
            # There is an available server: spawn a thread
            slave = self.slaveQueue[0]
            del self.slaveQueue[0]

            # Increment the count of active threads
            self.activeCount += 1
            thread = threading.Thread( target = self.slaveThread, \
                args=(slave,) )
            thread.start()

        self.lock.release()

    def __iter__( self ):
        """
        Returns an iterator over the (key, result) pairs
        returned by dequeue(). The iterator stops when there
        are no more results in the queue.
        """
        result = self.dequeue()
        while result != None:
            yield result
            result = self.dequeue()

    def dequeue( self ):
        """
        _dequeue_

        Returns a completed work item from the queue,
        or None if there are no more items. If there
        are worker threads still working, this will
        block until results are available.
        """

        result = None

        self.lock.acquire()

        while True:
            if len( self.resultsQueue ):
                # There is a result in the queue: take it
                result = self.resultsQueue[0]
                del self.resultsQueue[0]
                break
            elif self.activeCount == 0:
                # If there are no active threads, then there
                # are no results pending: quit
                result = None
                break
            else:
                assert( self.activeCount > 0 )
                # No result in the queue: wait for work
                self.lock.wait()

        self.lock.release()

        return result

    def slaveThread( self, slaveServer ):
        """
        _slaveThread_

        This thread executes the method for each work item.

        """
        #~ print "slave thread starting up"
        self.lock.acquire()
        assert( self.activeCount > 0 )

        while len( self.callQueue ) > 0:
            # Dequeue work
            key, parameters = self.callQueue[0]
            del self.callQueue[0]

            self.lock.release()
            #~ print "making a call..."
            results = slaveServer( *parameters )
            self.lock.acquire()

            self.resultsQueue.append( (key, results) )
            # Notify the main thread that results are available,
            # if it was waiting
            self.lock.notify()

        # This thread is all done: put it back in the queue
        #~ print "slave thread exiting"
        self.activeCount -= 1
        assert( self.activeCount >= 0 )
        self.slaveQueue.append( slaveServer )
        self.lock.release()



#!/usr/bin/env python
"""
_LoadFromID_

Oracle implementation of Jobs.LoadFromID.
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.LoadFromID import LoadFromID as LoadFromIDJobMySQL

class LoadFromID(LoadFromIDJobMySQL):
    sql = LoadFromIDJobMySQL.sql



"""
_Create_

Implementation of Createfor Oracle.

Inherit from CreateAgent, and add Oracle specific substitutions (e.g.
use trigger and sequence to mimic auto increment in MySQL.
"""




from WMCore.Agent.Database.CreateAgentBase import CreateAgentBase

class Create(CreateAgentBase):
    """
    Class to set up the WMBS schema in a MySQL database
    """
    sequenceTables = ["wm_components"]
    seqStartNum = 40

    def __init__(self, logger = None, dbi = None, params = None):
        """
        _init_

        Call the base class's constructor and create all necessary tables,
        constraints and inserts.
        """
        CreateAgentBase.__init__(self, logger, dbi, params)

        self.create["01wm_components"] = \
          """CREATE TABLE wm_components (
             id               INTEGER      NOT NULL,
             name             VARCHAR(255) NOT NULL,
             pid              INTEGER      NOT NULL,
             update_threshold INTEGER      NOT NULL,
             PRIMARY KEY (id),
             UNIQUE (name))"""


        for tableName in self.sequenceTables:
            seqname = '%s_SEQ' % tableName
            self.create["%s%s" % (self.seqStartNum, seqname)] = """
            CREATE SEQUENCE %s start with 1
            increment by 1 nomaxvalue cache 100""" % seqname

            triggerName = '%s_TRG' % tableName
            self.create["%s%s" % (self.seqStartNum, triggerName)] = """
                    CREATE TRIGGER %s
                        BEFORE INSERT ON %s
                        REFERENCING NEW AS newRow
                        FOR EACH ROW
                        BEGIN
                            SELECT %s.nextval INTO :newRow.id FROM dual;
                        END; """ % (triggerName, tableName, seqname)



#!/usr/bin/env python
"""
_Exists_

Oracle implementation of Files.Exists
"""

__all__ = []



from WMCore.WMBS.MySQL.Files.Exists import Exists as FilesExistsMySQL

class Exists(FilesExistsMySQL):
    sql = FilesExistsMySQL.sql



"""
_SetupCMSSWPset_

Create a CMSSW PSet suitable for running a WMAgent job.

"""
from builtins import next, object

import json
import logging
import os
import pickle
import socket

from PSetTweaks.PSetTweak import PSetTweak
from PSetTweaks.WMTweak import makeJobTweak, makeOutputTweak, makeTaskTweak, resizeResources
from Utils.Utilities import decodeBytesToUnicode, encodeUnicodeToBytes
from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig
from WMCore.WMRuntime.ScriptInterface import ScriptInterface
from WMCore.WMRuntime.Tools.Scram import Scram


def factory(module, name):
    """
    _factory_
    Function to return a dummy module name when a module
    is not available

    """

    class DummyClass(object):
        """
        _DummyClass_
        Dummy class to return when a cms class cannot be imported 

        """
        def __init__(self, module, name='', *args, **kwargs):
            self.__module = module
            self.__name = name
            self.__d = dict()

        def __setitem__(self, key, value):
            self.__d[key] = value

        def __getitem__(self, item):
            return self.__d[item]

        def __call__(self, *args, **kwargs):
            pass

        def __repr__(self):
            return "{module}.{name}".format(module=self.__module, name=self.__name)

    return DummyClass


class Unpickler(pickle.Unpickler):
    """
    _Unpickler_
    Use this when loading a PSet pickle
    and a specific module is not available (i.e.: FWCore modules)

    """
    def find_class(self, module, name):
        return factory(module, name)


class SetupCMSSWPset(ScriptInterface):
    """
    _SetupCMSSWPset_

    """

    def __init__(self, crabPSet=False):
        ScriptInterface.__init__(self)
        self.crabPSet = crabPSet
        self.process = None
        self.jobBag = None
        self.logger = logging.getLogger()
        self.tweak = PSetTweak()
        self.scram = None
        self.configPickle = "Pset.pkl"
        self.psetFile = None

    def createScramEnv(self):
        scramArchitecture = self.getScramVersion()
        cmsswVersion = self.getCmsswVersion()
        self.logger.info("Creating Scram environment with scram arch: %s and CMSSW version: %s",
            scramArchitecture,
            cmsswVersion)

        scram = Scram(
            version=cmsswVersion,
            directory=self.stepSpace.location,
            architecture=scramArchitecture,
            initialise=self.step.data.application.setup.softwareEnvironment
        )
        scram.project() # creates project area
        scram.runtime() # creates runtime environment

        return scram

    def scramRun(self, cmdArgs):
        """
        _scramRun_

        Run command inside scram environment

        """
        self.logger.info("ScramRun command args: %s", cmdArgs)
        if self.scram:
            retval = self.scram(command=cmdArgs)
            if retval > 0:
                msg = "Error running scram process. Error code: %s" % (retval)
                logging.error(msg)
                raise RuntimeError(msg)
        else:
            raise RuntimeError("Scram is not defined")

    def createProcess(self, scenario, funcName, funcArgs):
        """
        _createProcess_

        Create a Configuration.DataProcessing PSet.

        """

        procScript = "cmssw_wm_create_process.py"
        funcArgsJson = os.path.join(self.stepSpace.location, "process_funcArgs.json")

        if funcName not in ("merge", "repack"):
            funcArgs['scenario'] = scenario

        try:
            with open(funcArgsJson, 'w') as f:
                json.dump(funcArgs, f)
        except Exception as ex:
            msg = "Error writing out process funcArgs json."
            msg += "Type: {} and content: {}".format(type(funcArgs), funcArgs)
            self.logger.exception(msg)
            raise ex

        cmd = "%s --output_pkl %s --funcname %s --funcargs %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            funcName,
            funcArgsJson)

        if funcName == "merge":
            if getattr(self.jobBag, "useErrorDataset", False):
                cmd += " --useErrorDataset"

        self.scramRun(cmd)
        return

    def loadPSet(self):
        """
        _loadPSet_

        Load a PSet that was shipped with the job sandbox.
        Mock actual Pset values that depend on CMSSW, as these are
        handled externally.

        """
        self.logger.info("Working dir: %s", os.getcwd())
        # Pickle original pset configuration
        procScript = "edm_pset_pickler.py"
        cmd = "%s --input %s --output_pkl %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.psetFile),
            os.path.join(self.stepSpace.location, self.configPickle))
        self.scramRun(cmd)

        try:
            with open(os.path.join(self.stepSpace.location, self.configPickle), 'rb') as f:
                self.process = Unpickler(f).load()
        except ImportError as ex:
            msg = "Unable to import pset from %s:\n" % self.psetFile
            msg += str(ex)
            self.logger.error(msg)
            raise ex

        return

    def fixupProcess(self):
        """
        _fixupProcess_
        Look over the process object and make sure that all of the attributes
        that we expect to exist actually exist.
        """
        # Make sure that for each output module the following parameters exist
        # in the PSet returned from the framework:
        #   fileName
        #   logicalFileName
        #   dataset.dataTier
        #   dataset.filterName
        if hasattr(self.process, "outputModules"):
            outputModuleNames = list(self.process.outputModules) 
        elif hasattr(self.process, "outputModules_"):
            outputModuleNames = self.process.outputModules_()
        elif hasattr(self.process, "_Process__outputmodules"):
            outputModuleNames = list(self.process._Process__outputmodules)
        else:
            msg = "Error loading output modules from process"
            raise AttributeError(msg)

        for outMod in outputModuleNames:
            tweak = PSetTweak()
            self.logger.info("DEBUG output module = %s", outMod)
            tweak.addParameter("process.options", "customTypeCms.untracked.PSet()")
            tweak.addParameter("process.%s.dataset" % outMod, "customTypeCms.untracked.PSet(dataTier=cms.untracked.string(''), filterName=cms.untracked.string(''))")
            self.applyPsetTweak(tweak, skipIfSet=True, cleanupTweak=True)
            #tweak.addParameter("process.%s.dataset.dataTier" % outMod, "customTypeCms.untracked.string('')")
            #tweak.addParameter("process.%s.dataset.filterName" % outMod, "customTypeCms.untracked.string('')")
            tweak.addParameter("process.%s.fileName" % outMod, "customTypeCms.untracked.string('')")
            tweak.addParameter("process.%s.logicalFileName" % outMod, "customTypeCms.untracked.string('')")
            self.applyPsetTweak(tweak, skipIfSet=True)

        return

    def applyPsetTweak(self, psetTweak, skipIfSet=False, allowFailedTweaks=False, name='', cleanupTweak=False):
        """
        _applyPsetTweak_
        Apply a tweak to a pset process.
        Options:
          skipIfSet: Do not apply a tweak to a parameter that has a value set already.
          allowFailedTweaks: If the tweak of a parameter fails, do not abort and continue tweaking the rest.
          name: Extra string to add to the name of the json file that will be createed.
          cleanupTweak: Reset pset tweak object after applying all tweaks. Mostly used after using self.tweak
        """
        procScript = "edm_pset_tweak.py"
        psetTweakJson = os.path.join(self.stepSpace.location, "PSetTweak%s.json" % name)
        psetTweak.persist(psetTweakJson, formatting='simplejson')

        cmd = "%s --input_pkl %s --output_pkl %s --json %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle),
            psetTweakJson)
        if skipIfSet:
            cmd += " --skip_if_set"
        if allowFailedTweaks:
            cmd += " --allow_failed_tweaks"
        self.scramRun(cmd)

        if cleanupTweak:
            psetTweak.reset()

        return

    def handleSeeding(self):
        """
        _handleSeeding_

        Handle Random Seed settings for the job
        """
        seeding = getattr(self.jobBag, "seeding", None)
        seedJson = os.path.join(self.stepSpace.location, "reproducible_seed.json")
        self.logger.info("Job seeding set to: %s", seeding)
        procScript = "cmssw_handle_random_seeds.py"

        cmd = "%s --input_pkl %s --output_pkl %s --seeding %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle),
            seeding)

        if seeding == "ReproducibleSeeding":
            randService = self.jobBag.process.RandomNumberGeneratorService
            seedParams = {}
            for x in randService:
                parameter = "process.RandomNumberGeneratorService.%s.initialSeed" % x._internal_name
                seedParams[parameter] = x.initialSeed
            try:
                with open(seedJson, 'wb') as f:
                    json.dump(seedParams, f)
            except Exception as ex:
                self.logger.exception("Error writing out process funcArgs json:")
                raise ex
            cmd += " --reproducible_json %s" % (seedJson)

        self.scramRun(cmd)
        return


    def handlePerformanceSettings(self):
        """
        _handlePerformanceSettings_

        Install the standard performance report services
        """
        tweak = PSetTweak()
        # include the default performance report services
        if getattr(self.step.data.application.command, 'silentMemoryCheck', False):
            tweak.addParameter("process.SimpleMemoryCheck", "customTypeCms.Service('SimpleMemoryCheck', jobReportOutputOnly=cms.untracked.bool(True))")
        else:
            tweak.addParameter("process.SimpleMemoryCheck", "customTypeCms.Service('SimpleMemoryCheck')")

        tweak.addParameter("process.CPU", "customTypeCms.Service('CPU')")
        tweak.addParameter("process.Timing", "customTypeCms.Service('Timing')")
        self.applyPsetTweak(tweak)
        self.tweak.addParameter("process.Timing.summaryOnly", "customTypeCms.untracked(cms.bool(True))")

        return

    def makeThreadsStreamsTweak(self):
        """
        _makeThreadsStreamsTweak_

        Tweak threads and streams paraameters
        """
        origCores = int(getattr(self.step.data.application.multicore, 'numberOfCores', 1))
        eventStreams = int(getattr(self.step.data.application.multicore, 'eventStreams', 0))
        resources = {'cores': origCores}
        resizeResources(resources)
        numCores = resources['cores']
        if numCores != origCores:
            self.logger.info(
                "Resizing a job with nStreams != nCores. Setting nStreams = nCores. This may end badly.")
            eventStreams = 0

        tweak = PSetTweak()
        tweak.addParameter("process.options", "customTypeCms.untracked.PSet()")
        self.applyPsetTweak(tweak, skipIfSet=True)
        self.tweak.addParameter("process.options.numberOfThreads", "customTypeCms.untracked.uint32(%s)" % numCores)
        self.tweak.addParameter("process.options.numberOfStreams", "customTypeCms.untracked.uint32(%s)" % eventStreams)

        return

    def handleChainedProcessingTweak(self):
        """
        _handleChainedProcessing_

        When a job has multiple cmsRun steps, we need to tweak the
        subsequent PSet configuration such that it reads file in
        from the local disk (job sandbox).
        """
        self.logger.info("Handling chained processing job")

        # check the jobs input files
        inputFile = ("file:../%s/%s.root" % (self.step.data.input.inputStepName,
                                             self.step.data.input.inputOutputModule))

        self.tweak.addParameter('process.source.fileNames', "customTypeCms.untracked.vstring(%s)" % [inputFile])
        self.tweak.addParameter("process.maxEvents", "customTypeCms.untracked.PSet(input=cms.untracked.int32(-1))")
        self.logger.info("Chained PSet tweaked with maxEvents: -1, and fileNames: %s", inputFile)
        return

    def handlePileup(self):
        """
        _handlePileup_

        Handle pileup settings.
        There has been stored pileup configuration stored in a JSON file
        as a result of DBS querrying when running PileupFetcher,
        this method loads this configuration from sandbox and returns it
        as dictionary.
        The PileupFetcher was called by WorkQueue which creates job's sandbox
        and sandbox gets migrated to the worker node.

        External script iterates over all modules and over all pileup configuration types.
        The only considered types are "data" and "mc" (input to this method).
        If other pileup types are specified by the user, the method doesn't
        modify anything.
        The method considers only files which are present on this local PNN.
        The job will use only those, unless it was told to trust the PU site
        location (trustPUSitelists=True), in this case ALL the blocks/files
        will be added to the PSet and files will be read via AAA.
        Dataset, divided into blocks, may not have all blocks present on a
        particular PNN. However, all files belonging into a block will be
        present when reported by DBS.
        The structure of the pileupDict: PileupFetcher._queryDbsAndGetPileupConfig

        """
        # find out local site SE name
        siteConfig = loadSiteLocalConfig()
        PhEDExNodeName = siteConfig.localStageOut["phedex-node"]
        self.logger.info("Running on site '%s', local PNN: '%s'", siteConfig.siteName, PhEDExNodeName)
        jsonPileupConfig = os.path.join(self.stepSpace.location, "pileupconf.json")

        # Load pileup json
        try:
            with open(jsonPileupConfig) as jdata:
                pileupDict = json.load(jdata)
        except IOError:
            m = "Could not read pileup JSON configuration file: '%s'" % jsonPileupConfig
            raise RuntimeError(m)

        # Create a json with a list of files and events available
        # after dealing with PhEDEx/AAA logic
        newPileupDict = {}
        fileList = []
        eventsAvailable = 0
        for pileupType in self.step.data.pileup.listSections_():
            pileupType = decodeBytesToUnicode(pileupType)
            useAAA = True if getattr(self.jobBag, 'trustPUSitelists', False) else False
            self.logger.info("Pileup set to read data remotely: %s", useAAA)
            for blockName in sorted(pileupDict[pileupType].keys()):
                blockDict = pileupDict[pileupType][blockName]
                if PhEDExNodeName in blockDict["PhEDExNodeNames"] or useAAA:
                    eventsAvailable += int(blockDict.get('NumberOfEvents', 0))
                    for fileLFN in blockDict["FileList"]:
                        fileList.append(decodeBytesToUnicode(fileLFN))
            newPileupDict[pileupType] = {"eventsAvailable": eventsAvailable, "FileList": fileList}
        newJsonPileupConfig = os.path.join(self.stepSpace.location, "CMSSWPileupConfig.json")
        self.logger.info("Generating json for CMSSW pileup script")
        try:
            # If it's a python2 unicode, cmssw_handle_pileup will cast it to str
            with open(newJsonPileupConfig, 'w') as f:
                json.dump(newPileupDict, f)
        except Exception as ex:
            self.logger.exception("Error writing out process filelist json:")
            raise ex

        procScript = "cmssw_handle_pileup.py"
        cmd = "%s --input_pkl %s --output_pkl %s --pileup_dict %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle),
            newJsonPileupConfig)

        if getattr(self.jobBag, "skipPileupEvents", None):
            randomSeed = self.job['task']
            skipPileupEvents = self.jobBag.skipPileupEvents
            cmd += " --skip_pileup_events %s --random_seed %s" % (
                skipPileupEvents,
                randomSeed)
        self.scramRun(cmd)

        return

    def handleProducersNumberOfEvents(self):
        """
        _handleProducersNumberOfEvents_

        Some producer modules are initialized with a maximum number of events
        to be generated, usually based on the process.maxEvents.input attribute
        but after that is tweaked the producers number of events need to
        be fixed as well. This method takes care of that.
        """

        procScript = "cmssw_handle_nEvents.py"
        cmd = "%s --input_pkl %s --output_pkl %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle))
        self.scramRun(cmd)

        return


    def handleDQMFileSaver(self):
        """
        _handleDQMFileSaver_

        Harvesting jobs have the dqmFileSaver EDAnalyzer that must
        be tweaked with the dataset name in order to store it
        properly in the DQMGUI, others tweaks can be added as well
        """

        runIsComplete = getattr(self.jobBag, "runIsComplete", False)
        multiRun = getattr(self.jobBag, "multiRun", False)
        runLimits = getattr(self.jobBag, "runLimits", "")
        self.logger.info("DQMFileSaver set to multiRun: %s, runIsComplete: %s, runLimits: %s",
                         multiRun, runIsComplete, runLimits)

        procScript = "cmssw_handle_dqm_filesaver.py"

        cmd = "%s --input_pkl %s --output_pkl %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle))

        if hasattr(self.step.data.application.configuration, "pickledarguments"):
            pklArgs = encodeUnicodeToBytes(self.step.data.application.configuration.pickledarguments)
            args = pickle.loads(pklArgs)
            datasetName = args.get('datasetName', None)
        if datasetName:
            cmd += " --datasetName %s" % (datasetName)
        if multiRun and runLimits:
            cmd += " --multiRun --runLimits=%s" % (runLimits)
        if runIsComplete:
            cmd += " --runIsComplete"
        self.scramRun(cmd)

        return

    def handleLHEInput(self):
        """
        _handleLHEInput_

        Enable lazy-download for jobs reading LHE articles from CERN, such
        that these jobs can read data remotely
        """

        if getattr(self.jobBag, "lheInputFiles", False):
            self.logger.info("Enabling 'lazy-download' for lheInputFiles job")
            self._enableLazyDownload()

        return

    def handleRepackSettings(self):
        """
        _handleRepackSettings_

        Repacking small events is super inefficient reading directly from EOS.
        """
        self.logger.info("Hardcoding read/cache strategies for repack")
        self._enableLazyDownload()
        return

    def _enableLazyDownload(self):
        """
        _enableLazyDownload_

        Set things to read data remotely
        """
        procScript = "cmssw_enable_lazy_download.py"
        cmd = "%s --input_pkl %s --output_pkl %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle))
        self.scramRun(cmd)

        return

    def handleSingleCoreOverride(self):
        """
        _handleSingleCoreOverride_

        Make sure job only uses one core and one stream in CMSSW
        """
        try:
            if int(self.step.data.application.multicore.numberOfCores) > 1:
                self.step.data.application.multicore.numberOfCores = 1
        except AttributeError:
            pass

        try:
            if int(self.step.data.application.multicore.eventStreams) > 0:
                self.step.data.application.multicore.eventStreams = 0
        except AttributeError:
            pass

        return

    def handleSpecialCERNMergeSettings(self, funcName):
        """
        _handleSpecialCERNMergeSettings_

        CERN has a 30ms latency between Meyrin and Wigner, which kills merge performance
        Enable lazy-download for fastCloning for all CMSSW_7_5 jobs (currently off)
        Enable lazy-download for all merge jobs
        """
        if self.getCmsswVersion().startswith("CMSSW_7_5") and False:
            self.logger.info("Using fastCloning/lazydownload")
            self._enableLazyDownload()
        elif funcName == "merge":
            self.logger.info("Using lazydownload")
            self._enableLazyDownload()

        return

    def handleCondorStatusService(self):
        """
        _handleCondorStatusService_

        Enable CondorStatusService for CMSSW releases that support it.
        """
        procScript = "cmssw_handle_condor_status_service.py"
        cmd = "%s --input_pkl %s --output_pkl %s --name %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle),
            self.step.data._internal_name)
        self.scramRun(cmd)

        return

    def handleEnforceGUIDInFileName(self, secondaryInput=None):
        """
        _handleEnforceGUIDInFileName_

        Enable enforceGUIDInFileName for CMSSW releases that support it.
        """
        # skip it for CRAB jobs
        if self.crabPSet:
            return

        if secondaryInput:
            inputSource = secondaryInput
            self.logger.info("Evaluating enforceGUIDInFileName parameter for secondary input data.")
        else:
            inputSource = self.process.source

        if hasattr(inputSource, "type_"):
            inputSourceType = inputSource.type_()
        elif hasattr(inputSource, "_TypedParameterizable__type"):
            inputSourceType = inputSource._TypedParameterizable__type
        else:
            msg = "Source type could not be determined."
            self.logger.error(msg)
            raise AttributeError(msg)

        # only enable if source is PoolSource or EmbeddedRootSource
        if inputSourceType not in ["PoolSource", "EmbeddedRootSource"]:
            self.logger.info("Not evaluating enforceGUIDInFileName parameter for process source %s",
                             inputSourceType)
            return

        procScript = "cmssw_enforce_guid_in_filename.py"
        cmd = "%s --input_pkl %s --output_pkl %s --input_source %s" % (
            procScript,
            os.path.join(self.stepSpace.location, self.configPickle),
            os.path.join(self.stepSpace.location, self.configPickle),
            inputSourceType)
        self.scramRun(cmd)

        return

    def getCmsswVersion(self):
        """
        _getCmsswVersion_

        Return a string representing the CMSSW version to be used.
        """
        if not self.crabPSet:
            return self.step.data.application.setup.cmsswVersion

        # CRAB3 needs to use an environment var to get the version
        return os.environ.get("CMSSW_VERSION", "")


    def getScramVersion(self, allSteps=False):
        """
        _getScramVersion_

        Return a string representing the first Scram version to be used (or all)
        """
        if not self.crabPSet:
            scramArch = self.step.data.application.setup.scramArch
            if allSteps:
                return scramArch
            else:
                if isinstance(scramArch, list):
                    return next(iter(scramArch or []), None)

        # CRAB3 needs to use an environment var to get the version
        return os.environ.get("SCRAM_ARCH", "")


    def __call__(self):
        """
        _call_

        Examine the step configuration and construct a PSet from that.

        """
        self.logger.info("Executing SetupCMSSWPSet...")
        self.jobBag = self.job.getBaggage()
        self.configPickle = getattr(self.step.data.application.command, "configurationPickle", "PSet.pkl")
        self.psetFile = getattr(self.step.data.application.command, "configuration", "PSet.py")
        self.scram = self.createScramEnv()

        scenario = getattr(self.step.data.application.configuration, "scenario", None)
        funcName = getattr(self.step.data.application.configuration, "function", None)
        if scenario is not None and scenario != "":
            self.logger.info("Setting up job scenario/process")
            if getattr(self.step.data.application.configuration, "pickledarguments", None) is not None:
                pklArgs = encodeUnicodeToBytes(self.step.data.application.configuration.pickledarguments)
                funcArgs = pickle.loads(pklArgs)
            else:
                funcArgs = {}

            # Create process
            try:
                self.createProcess(scenario, funcName, funcArgs)
            except Exception as ex:
                self.logger.exception("Error creating process for Config/DataProcessing:")
                raise ex
            # Now, load the new picked process
            try:
                with open(os.path.join(self.stepSpace.location, self.configPickle), 'rb') as f:
                    self.process = Unpickler(f).load()
            except ImportError as ex:
                msg = "Unable to import pset from %s:\n" % self.psetFile
                msg += str(ex)
                self.logger.error(msg)
                raise ex

            if funcName == "repack":
                self.handleRepackSettings()

            if funcName in ["merge", "alcaHarvesting"]:
                self.handleSingleCoreOverride()

            if socket.getfqdn().endswith("cern.ch"):
                self.handleSpecialCERNMergeSettings(funcName)
        else:
            self.logger.info("DEBUG: Now in the none scenario to load PSET")
            try:
                self.loadPSet()
            except Exception as ex:
                self.logger.exception("Error loading PSet:")
                raise ex

        # Check process.source exists
        self.logger.info("Debug: Self.process")
        self.logger.info(dir(self.process))
        if getattr(self.process, "source", None) is None and getattr(self.process, "_Process__source", None) is None:
            msg = "Error in CMSSW PSet: process is missing attribute 'source'"
            msg += " or process.source is defined with None value."
            self.logger.error(msg)
            raise RuntimeError(msg)

        self.handleCondorStatusService()
        self.fixupProcess()

        # In case of CRAB3, the number of threads in the PSet should not be overridden
        if not self.crabPSet:
            try:
                self.makeThreadsStreamsTweak()
            except AttributeError as ex:
                self.logger.error("Failed to override numberOfThreads: %s", str(ex))

        # Apply task level tweaks
        makeTaskTweak(self.step.data, self.tweak)
        self.applyPsetTweak(self.tweak, cleanupTweak=True)

        # Check if chained processing is enabled
        # If not - apply the per job tweaks
        # If so - create an override TFC (like done in PA) and then modify thePSet accordingly
        if hasattr(self.step.data.input, "chainedProcessing") and self.step.data.input.chainedProcessing:
            self.logger.info("Handling Chain processing tweaks")
            self.handleChainedProcessingTweak()
        else:
            self.logger.info("Creating job level tweaks")
            makeJobTweak(self.job, self.tweak)
        self.applyPsetTweak(self.tweak, cleanupTweak=True)

        # check for pileup settings presence, pileup support implementation
        # and if enabled, process pileup configuration / settings
        if hasattr(self.step.data, "pileup"):
            self.handlePileup()

        # Apply per output module PSet Tweaks
        self.logger.info("Output module section")
        cmsswStep = self.step.getTypeHelper()
        for om in cmsswStep.listOutputModules():
            mod = cmsswStep.getOutputModule(om)
            modName = mod.getInternalName()

            if funcName == 'merge':
                # Do not use both Merged output label unless useErrorDataset is False
                # Do not use both MergedError output label unless useErrorDataset is True 
                useErrorDataset = getattr(self.jobBag, "useErrorDataset", False)

                if useErrorDataset and modName != 'MergedError':
                    continue
                if not useErrorDataset and modName == 'MergedError':
                    continue

            makeOutputTweak(mod, self.job, self.tweak)
        # allow failed tweaks in this case, to replicate the previous implementation, where it would ignore 
        # and continue if it found an output module that  doesn't exist and don't want in the pset like: process.Sqlite
        self.applyPsetTweak(self.tweak, allowFailedTweaks=True, cleanupTweak=True)

        # revlimiter for testing
        if getattr(self.step.data.application.command, "oneEventMode", False):
            self.tweak.addParameter('process.maxEvents.input', "customTypeCms.untracked.int32(1)")

        # check for random seeds and the method of seeding which is in the job baggage
        self.handleSeeding()

        # make sure default parametersets for perf reports are installed
        self.handlePerformanceSettings()

        # fixup the dqmFileSaver
        self.handleDQMFileSaver()

        # tweak for jobs reading LHE articles from CERN
        self.handleLHEInput()

        # tweak jobs for enforceGUIDInFileName
        self.handleEnforceGUIDInFileName()

        # Check if we accept skipping bad files
        if hasattr(self.step.data.application.configuration, "skipBadFiles"):
            self.tweak.addParameter("process.source.skipBadFiles",
                "customTypeCms.untracked.bool(%s)" % self.step.data.application.configuration.skipBadFiles)

        # Apply events per lumi section if available
        if hasattr(self.step.data.application.configuration, "eventsPerLumi"):
            self.tweak.addParameter("process.source.numberEventsInLuminosityBlock",
                "customTypeCms.untracked.uint32(%s)" % self.step.data.application.configuration.eventsPerLumi)

        # limit run time if desired
        if hasattr(self.step.data.application.configuration, "maxSecondsUntilRampdown"):
            self.tweak.addParameter("process.maxSecondsUntilRampdown.input",
                "customTypeCms.untracked.PSet(input=cms.untracked.int32(%s))" % self.step.data.application.configuration.maxSecondsUntilRampdown)

        # accept an overridden TFC from the step
        if hasattr(self.step.data.application, 'overrideCatalog'):
            self.logger.info("Found a TFC override: %s", self.step.data.application.overrideCatalog)
            self.tweak.addParameter("process.source.overrideCatalog",
                "customTypeCms.untracked.string('%s')" % self.step.data.application.overrideCatalog)

        configFile = self.step.data.application.command.configuration
        workingDir = self.stepSpace.location
        try:
            self.applyPsetTweak(self.tweak)

            with open("%s/%s" % (workingDir, configFile), 'w') as handle:
                handle.write("import FWCore.ParameterSet.Config as cms\n")
                handle.write("import pickle\n")
                handle.write("with open('%s', 'rb') as handle:\n" % self.configPickle)
                handle.write("    process = pickle.load(handle)\n")
        except Exception as ex:
            self.logger.exception("Error writing out PSet:")
            raise ex

        # check for event numbers in the producers
        self.handleProducersNumberOfEvents()

        self.logger.info("CMSSW PSet setup completed!")

        return 0



# -*- coding: utf-8 -*-
"""
_GetRunningSlots_

Oracle implementation of Locations.GetRunningSlots
Created on Mon Jun 18 13:49:39 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Locations.GetRunningSlots import GetRunningSlots as MySQLGetRunningSlots

class GetRunningSlots(MySQLGetRunningSlots):
    """
    Same as MySQL version

    """
    pass



#!/usr/bin/env python
"""
_Parentage_

Oracle implementation of Fileset.Parentage

"""
__all__ = []



from WMCore.WMBS.MySQL.Fileset.Parentage import Parentage as FilesetParentageMySQL

class Parentage(FilesetParentageMySQL):
    sql = FilesetParentageMySQL.sql



"""
_CreateWMBS_

Implementation of CreateWMBS for MySQL.

Inherit from CreateWMBSBase, and add MySQL specific substitutions (e.g. add
INNODB) and specific creates (e.g. for time stamp and enum fields).
"""

from WMCore.WMBS.CreateWMBSBase import CreateWMBSBase


class Create(CreateWMBSBase):
    """
    Class to set up the WMBS schema in a MySQL database
    """

    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the base class's constructor and create all necessary tables,
        constraints and inserts.
        """
        CreateWMBSBase.__init__(self, logger, dbi, params)

        self.create["01wmbs_fileset"] = \
            """CREATE TABLE wmbs_fileset (
               id          INTEGER      PRIMARY KEY AUTO_INCREMENT,
               name        VARCHAR(1250) NOT NULL,
               open        INT(1)       NOT NULL DEFAULT 0,
               last_update INT(11)      NOT NULL,
               UNIQUE (name))"""

        self.create["03wmbs_fileset_files"] = \
            """CREATE TABLE wmbs_fileset_files (
               fileid      INT(11)   NOT NULL,
               fileset     INT(11)   NOT NULL,
               insert_time INT(11)   NOT NULL,
               UNIQUE(fileid, fileset),
               FOREIGN KEY(fileset) REFERENCES wmbs_fileset(id)
                 ON DELETE CASCADE,
               FOREIGN KEY(fileid)  REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE)"""

        self.create["13wmbs_jobgroup"] = \
            """CREATE TABLE wmbs_jobgroup (
               id           INT(11)    NOT NULL AUTO_INCREMENT,
               subscription INT(11)    NOT NULL,
               guid         VARCHAR(255),
               output       INT(11),
               last_update  INT(11)    NOT NULL,
               location     INT(11),
               PRIMARY KEY (id),
               FOREIGN KEY (subscription) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (output) REFERENCES wmbs_fileset(id)
                      ON DELETE CASCADE)"""

        self.constraints["uniquewfname"] = \
            "CREATE UNIQUE INDEX uniq_wf_name on wmbs_workflow (name, task)"

        self.constraints["uniquefilerunlumi"] = \
            """CREATE UNIQUE INDEX uniq_wmbs_file_run_lumi on
               wmbs_file_runlumi_map (fileid, run, lumi)"""

    def execute(self, conn=None, transaction=None):
        for i in self.create:
            self.create[i] += " ENGINE=InnoDB ROW_FORMAT=DYNAMIC"

        return CreateWMBSBase.execute(self, conn, transaction)



#!/usr/bin/env python
"""
_JobsByState_

Monitoring DAO classes for Jobs in WMBS
"""

from WMCore.Database.DBFormatter import DBFormatter


class JobCountByState(DBFormatter):
    sql = """SELECT wmbs_job_state.name AS job_state, count(wmbs_job.state) AS job_count
               FROM wmbs_job_state
               LEFT OUTER JOIN wmbs_job ON wmbs_job.state=wmbs_job_state.id
             GROUP BY wmbs_job_state.name"""

    def formatDict(self, results):
        """
        _formatDict_

        Format the results in a dict keyed by the state name
        """
        formattedResults = DBFormatter.formatDict(self, results)

        dictResult = {}
        for formattedResult in formattedResults:
            dictResult[formattedResult['job_state']] = int(formattedResult['job_count'])

        return dictResult

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        _execute_

        Execute the SQL.
        """
        result = self.dbi.processData(self.sql, conn=conn, transaction=transaction)

        return self.formatDict(result)



"""
ArchiveDataPoller
"""

from __future__ import (division, print_function)
import logging
import traceback
from Utils.IteratorTools import grouper
from Utils.Timers import timeFunction
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Services.WMArchive.DataMap import createArchiverDoc
from WMCore.Services.WMArchive.WMArchive import WMArchive
from WMCore.Services.FWJRDB.FWJRDBAPI import FWJRDBAPI


class ArchiveDataPoller(BaseWorkerThread):
    """
    Gather fwjr data and update to archiver\
    """

    def __init__(self, config):
        """
        initialize properties specified from config
        """
        BaseWorkerThread.__init__(self)
        self.config = config

    def setup(self, parameters):
        """
        set db connection(couchdb, wmbs) to prepare to gather information
        """
        baseURL = self.config.JobStateMachine.couchurl
        dbname = "%s/fwjrs" % getattr(self.config.JobStateMachine, "couchDBName")

        self.fwjrAPI = FWJRDBAPI(baseURL, dbname)
        self.wmarchiver = WMArchive(self.config.ArchiveDataReporter.WMArchiveURL)
        self.numDocsRetrievePerPolling = getattr(self.config.ArchiveDataReporter, "numDocsRetrievePerPolling", 1000)
        self.numDocsUploadPerCall = getattr(self.config.ArchiveDataReporter, "numDocsUploadPerCall", 200)

    @timeFunction
    def algorithm(self, parameters):
        """
        get information from wmbs, workqueue and local couch
        """
        try:
            data = self.fwjrAPI.getFWJRByArchiveStatus('ready', limit=self.numDocsRetrievePerPolling)['rows']
            logging.info("Found %i not archived documents from FWRJ db to upload to WMArchive.", len(data))

            for slicedData in grouper(data, self.numDocsUploadPerCall):
                jobIDs = []
                archiveDocs = []
                for job in slicedData:
                    doc = createArchiverDoc(job)
                    archiveDocs.append(doc)
                    jobIDs.append(job["id"])

                response = self.wmarchiver.archiveData(archiveDocs)

                # Partial success is not allowed either all the insert is successful or none is
                if response[0]['status'] == "ok" and len(response[0]['ids']) == len(jobIDs):
                    archiveIDs = response[0]['ids']
                    for docID in jobIDs:
                        self.fwjrAPI.updateArchiveUploadedStatus(docID)
                    logging.info("...successfully uploaded %d docs", len(jobIDs))
                    logging.debug("JobIDs uploaded: %s", jobIDs)
                    logging.debug("Archived IDs returned: %s", archiveIDs)
                else:
                    logging.warning("Upload failed and it will be retried in the next cycle: %s: %s.",
                                    response[0]['status'], response[0]['reason'])
                    logging.debug("failed JobIds %s", jobIDs)
        except Exception as ex:
            logging.error("Error occurred, will retry later:")
            logging.error(str(ex))
            logging.error("Trace back: \n%s", traceback.format_exc())



#!/usr/bin/env python
#pylint: disable=E1101,E1103,C0103,R0902

"""
Performs bulk DBS File(s) insertion by :
        reading the FJR received in payload
        buffering in the database
        if buffer has hit the configured limit
"""
from __future__ import print_function





import logging
import threading


from WMCore.Agent.Harness import Harness
from WMCore.WMFactory import WMFactory

from WMComponent.DBS3Buffer.DBSUploadPoller import DBSUploadPoller




class DBS3Upload(Harness):
    """
    Load the poller thread

    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.pollTime = 1
        print("DBS3Upload.__init__")

    def preInitialization(self):
        print("DBS3Upload.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()

        pollInterval = self.config.DBS3Upload.pollInterval
        logging.info("Setting poll interval to %s seconds" % pollInterval)
        myThread.workerThreadManager.addWorker(DBSUploadPoller(self.config), pollInterval)

        return



#!/usr/bin/env python
"""
_JobStatusByWorkflowAndSite_

MySQL implementation for loading a job by scheduler status
"""


from WMCore.Database.DBFormatter import DBFormatter
from WMCore.BossAir.Plugins.BasePlugin import BasePlugin

class JobStatusByWorkflowAndSite(DBFormatter):
    """
    _LoadForMonitoring_

    Load all jobs with a certain scheduler status including
    all the joined information.
    """

    #TODO: not sure why location is left outer join instead of inner join
    sql = """SELECT wwf.name as workflow, count(rj.wmbs_id) AS num_jobs,
                    st.name AS status, wl.cms_name AS site,
                    wl.plugin AS plugin, wu.cert_dn AS owner
               FROM bl_runjob rj
               INNER JOIN wmbs_users wu ON wu.id = rj.user_id
               INNER JOIN bl_status st ON rj.sched_status = st.id
               INNER JOIN wmbs_job wj ON wj.id = rj.wmbs_id
               INNER JOIN wmbs_jobgroup wjg ON wjg.id = wj.jobgroup
               INNER JOIN wmbs_subscription ws ON ws.id = wjg.subscription
               INNER JOIN wmbs_workflow wwf ON wwf.id = ws.workflow
               LEFT OUTER JOIN wmbs_location wl ON wl.id = wj.location
               WHERE rj.status = :complete
               GROUP BY wwf.name, plugin, st.name, wl.cms_name, wu.cert_dn
    """
    #TODO this needs to move to BasePlugin replacing the global state: check with Matt
    MONITOR_STATE_MAP = {'Pending': 'submitted_pending',
                         'Running': 'submitted_running',
                         'Complete': 'submitted_running',
                         'Error': 'submitted_running'}

    def mappedStatusFormat(self, results):
        """
        convert each indiviual batch system plugin status to a common status.
        Warning: This assumes all the plugin are under WMCore/BossAir/Plugins/
        and its module name and class name should be the same.

        __import__ doesn't reload if the module exist.
        (so doesn't need to keep track what is already imported.
        The performance difference should be small. If desired, maintain
        the cache to keep track imported plugIns
        TODO: need to reduce state only pending_batch and running_batch
        """
        commonStates = {}
        for data in results:
            module = __import__("WMCore.BossAir.Plugins.%s" % data['plugin'],
                                globals(), locals(), [data['plugin']])
            plugIn = getattr(module, data['plugin'])
            state = self.MONITOR_STATE_MAP[plugIn.stateMap().get(data['status'])]
            if data['workflow'] not in commonStates:
                commonStates[data['workflow']] = {}

            commonStates[data['workflow']].setdefault(state, {})
            commonStates[data['workflow']][state].setdefault(data['site'], 0)
            commonStates[data['workflow']][state][data['site']] += data['num_jobs']
        return commonStates

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        Load all jobs either running or not (running by default)
        """
        complete = '1'
        binds = {'complete': complete}


        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return self.mappedStatusFormat(self.formatDict(result))



#!/usr/bin/env python
"""
_LoadByWMBSID_

MySQL implementation for loading a job by WMBS info
"""

from WMCore.Database.DBFormatter import DBFormatter


class LoadByWMBSID(DBFormatter):
    """
    _LoadByWMBSID_

    Load all jobs in full by WMBS ID and retry count
    """

    sql = """SELECT rj.wmbs_id AS jobid, rj.grid_id AS gridid, rj.bulk_id AS bulkid,
               st.name AS status, rj.retry_count as retry_count, rj.id AS id,
               rj.status_time as status_time, wu.cert_dn AS userdn,
               wu.group_name AS usergroup, wu.role_name AS userrole,
               wl.plugin AS plugin, wj.cache_dir AS cache_dir
               FROM bl_runjob rj
               INNER JOIN bl_status st ON rj.sched_status = st.id
               LEFT OUTER JOIN wmbs_users wu ON wu.id = rj.user_id
               INNER JOIN wmbs_job wj ON wj.id = rj.wmbs_id
               INNER JOIN wmbs_location wl ON wl.id = wj.location
               WHERE rj.wmbs_id = :id AND rj.retry_count = :retry_count"""

    def execute(self, jobs, conn=None, transaction=False):
        """
        _execute_

        Load jobs in full via WMBS ID and retry count
        Only useful when killing outside of tracker/submitter loop
        where you don't have any bossAir info
        """

        if len(jobs) < 1:
            # No jobs to run
            return

        binds = []
        for job in jobs:
            binds.append({'id': job['id'], 'retry_count': job['retry_count']})

        result = self.dbi.processData(self.sql, binds, conn=conn,
                                      transaction=transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_DBS3Buffer.CreateBlocks_

Create new block in dbsbuffer_block

"""

from WMCore.Database.DBFormatter import DBFormatter

class CreateBlocks(DBFormatter):

    sql = """INSERT INTO dbsbuffer_block
             (dataset_id, blockname, location, status, create_time)
             SELECT (SELECT id from dbsbuffer_dataset WHERE path = :dataset),
                    :block,
                    (SELECT id FROM dbsbuffer_location WHERE pnn = :location),
                    :status,
                    :time
             FROM DUAL
             """

    def execute(self, blocks, conn = None, transaction = False):

        bindVars = []

        for block in blocks:

            bindVars.append( { 'dataset' : block.getDataset(),
                               'block' : block.getName(),
                               'location' : block.getLocation(),
                               'status' : block.status,
                               'time' : block.getStartTime() } )

        self.dbi.processData(self.sql, bindVars, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
_CompleteFiles_

Oracle implementation of Subscription.CompleteFiles
"""

from WMCore.WMBS.MySQL.Subscriptions.CompleteFiles import CompleteFiles \
     as CompleteFilesMySQL

class CompleteFiles(CompleteFilesMySQL):
    sql = """INSERT INTO wmbs_sub_files_complete (subscription, fileid)
               SELECT :subscription, :fileid FROM dual WHERE NOT EXISTS
                 (SELECT fileid FROM wmbs_sub_files_complete
                    WHERE fileid = :fileid AND subscription = :subscription)"""



#!/usr/bin/env python
"""
_ListSubTypes_

Oracle implementation of Monitoring.ListSubTypes
"""




from WMCore.WMBS.MySQL.Monitoring.ListSubTypes import ListSubTypes \
    as ListSubTypesMySQL

class ListSubTypes(ListSubTypesMySQL):
    pass



#!/usr/bin/env python
"""
_CPImpl_

Implementation of StageOutImpl interface for plain cp

"""
from __future__ import division

import os
import errno
import shutil

from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.Storage.StageOutError import StageOutFailure


class CPImpl(StageOutImplV2):
    """
    _CPImpl_

    Implement interface for plain cp command

    """

    def createOutputDirectory(self, targetPFN):
        targetdir = os.path.dirname(targetPFN)
        if not os.path.isdir(targetdir):
            try:
                os.makedirs(targetdir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise
        return

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum ):
        self.createOutputDirectory(toPfn)
        shutil.copy(fromPfn, toPfn)
        if os.path.getsize(fromPfn) != os.path.getsize(toPfn):
            raise StageOutFailure("Invalid file size")
        return toPfn

    def doDelete(self, pfn, pnn, command, options, protocol):
        os.remove(pfn)
        return



#!/usr/bin/env python
"""
_GetCache_

Oracle implementation of Jobs.GetCache
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.GetCache import GetCache as MySQLGetCache

class GetCache(MySQLGetCache):
    """
    Identical to MySQL version for now

    """



#!/usr/bin/env python
"""
_LoadBlocksByDAS_

MySQL implementation of LoadBlocksByDAS
"""




from WMCore.Database.DBFormatter import DBFormatter

class LoadBlocksByDAS(DBFormatter):
    sql = """SELECT DISTINCT dbb.blockname as blockname, dbb.create_time as create_time,
                (SELECT COUNT(*) FROM dbsbuffer_file dbf WHERE dbf.block_id = dbb.id) AS nFiles,
                (SELECT SUM(dbf2.filesize) FROM dbsbuffer_file dbf2 WHERE dbf2.block_id = dbb.id) AS blocksize,
                dbl.pnn AS location
                FROM dbsbuffer_block dbb
                INNER JOIN dbsbuffer_file dbf3 ON dbf3.block_id = dbb.id
                INNER JOIN dbsbuffer_location dbl ON dbl.id = dbb.location
                WHERE dbb.status = 'Open'
                AND dbf3.dataset_algo = :das"""


    def format(self, result):
        tmpList = self.formatDict(result)
        blockList = []
        for tmp in tmpList:
            final = {}
            final['Name']          = tmp['blockname']
            final['CreationDate']  = tmp['create_time']
            final['NumberOfFiles'] = tmp['nfiles']
            final['BlockSize']     = float(tmp['blocksize'])
            final['location']      = tmp['location']
            final['newFiles']      = []
            final['insertedFiles'] = []
            final['open']          = 1
            blockList.append(final)

        return blockList


    def execute(self, das, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {'das': das},
                         conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_CMSSW_

Basic Emulator for CMSSW Step

"""
from WMCore.WMSpec.Steps.Emulator import Emulator
from WMCore.FwkJobReport.ReportEmu import ReportEmu
from WMCore.WMSpec.WMStep import WMStepHelper

class CMSSW(Emulator):
    """
    _CMSSW_

    Emulate the execution of a CMSSW Step

    """

    def execute(self):
        """
        _emulate_


        """
        self.step.section_("execution")
        self.step.execution.exitStatus = 0
        self.step.emulator.emulatedBy = str(self.__class__.__name__)

        helper = WMStepHelper(self.step)
        cmsswStep = helper.getTypeHelper()
        reportMaker = ReportEmu(Job = self.job, WMStep = cmsswStep)
        self.executor.report = reportMaker()

        # TODO: touch physical file names to make it look like the files are actually
        #       produced during emulation

    def pre(self):
        '''
            _emulatePre_

            Emulates the pre-execution step
        '''
        return self.executor.pre()

    def post(self):
        '''
            _emulatePost_

            Emulates the post-execution step
        '''
        return self.executor.post()



#!/usr/bin/env python
"""
_Proxy_
Wrap gLite proxy commands.
"""

from __future__ import division
from builtins import filter, str, range

import contextlib
import copy
import os
import re
import subprocess
import time
from datetime import datetime
from hashlib import sha1

from WMCore.Credential.Credential import Credential
from WMCore.WMException import WMException
from Utils.PythonVersion import PY3
from Utils.Utilities import decodeBytesToUnicode, encodeUnicodeToBytes


def execute_command(command, logger, timeout, redirect=True):
    """
    _execute_command_
    Function to manage commands.
    """

    stdout, stderr, rc = None, None, 99999
    if redirect:
        proc = subprocess.Popen(
            command, shell=True, cwd=os.environ['PWD'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )
    else:
        proc = subprocess.Popen(command, shell=True, cwd=os.environ['PWD'])

    t_beginning = time.time()
    while True:
        if proc.poll() is not None:
            break
        seconds_passed = time.time() - t_beginning
        if timeout and seconds_passed > timeout:
            proc.terminate()
            logger.error('Timeout in %s execution.' % command)
            return stdout, rc

        time.sleep(0.1)

    stdout, stderr = proc.communicate()
    stdout = decodeBytesToUnicode(stdout) if PY3 else stdout
    stderr = decodeBytesToUnicode(stderr) if PY3 else stderr
    rc = proc.returncode

    logger.debug('Executing : \n command : %s\n output : %s\n error: %s\n retcode : %s' % (command, stdout, stderr, rc))

    return stdout, stderr, rc


def destroyListCred(credNameList=None, credTimeleftList=None, logger=None, timeout=0):
    """
    _destroyListCred_
    Get list of credential name and their timelefts to destroy the one
    with timeleft = 0 from myproxy.
    """
    credNameList = credNameList or []
    credTimeleftList = credTimeleftList or {}
    cleanCredCmdList = []

    for credIdx in range(len(credNameList)):
        hours, minutes, seconds = credTimeleftList[credIdx]
        timeleft = int(hours) * 3600 + int(minutes) * 60 + int(seconds)
        if timeleft == 0:
            cleanupCmd = "myproxy-destroy -d -k %s" % (credNameList[credIdx])
            cleanCredCmdList.append(cleanupCmd)

    cleanCredCmd = " && ".join(cleanCredCmdList)
    if len(cleanCredCmd) > 0:
        execute_command(cleanCredCmd, logger, timeout)

    return


# TODO not used anymore. #3810 deletes lasts unused dependencies in the client
class CredentialException(WMException):
    """
    Credential exceptions should be defined in this class.
    """
    pass


@contextlib.contextmanager
def myProxyEnvironment(userDN, serverCert, serverKey, myproxySrv, proxyDir, logger):
    """
    Allows us to user a context manager within which a MyProxy delegated proxy
    is set to X509_USER_PROXY and things are restored on exit
    """

    originalEnvironment = copy.deepcopy(os.environ)
    try:
        args = {}
        args['server_cert'] = serverCert
        args['server_key'] = serverKey
        args['myProxySvr'] = myproxySrv
        args['credServerPath'] = proxyDir
        args['logger'] = logger
        proxy = Proxy(args=args)

        proxy.userDN = userDN
        filename = proxy.logonRenewMyProxy()
        os.environ['X509_USER_PROXY'] = filename

        # host certs can be taken first, get rid of them
        deleteKeys = ['X509_USER_CERT', 'X509_USER_KEY', 'X509_HOST_CERT', 'X509_HOST_KEY']
        for key in deleteKeys:
            if key in os.environ:
                del os.environ[key]
        yield filename
    finally:
        os.environ = originalEnvironment


class Proxy(Credential):
    """
    Basic class to handle user Proxy
    """

    def __init__(self, args):
        """
        __init__
        Initialize proxy object.
        """
        Credential.__init__(self, args)

        # Set the default commands execution timeout to 20 mn
        self.commandTimeout = args.get("ServiceContactTimeout", 1200)
        self.myproxyServer = args.get("myProxySvr", 'myproxy.cern.ch')
        self.serverDN = args.get("serverDN", '')
        self.userDN = args.get("userDN", '')
        self.proxyValidity = args.get("proxyValidity", '')  # lenght of the proxy
        self.myproxyValidity = args.get("myproxyValidity", '168:00')  # lenght of the myproxy
        self.myproxyMinTime = args.get("myproxyMinTime", 4)  # threshold used in checkProxy
        self.myproxyAccount = args.get("myproxyAccount", "")  # to be used when computing myproxy account (-l option)
        self.rfcCompliant = args.get("rfcCompliant", True)  # to be used when computing myproxy account (-l option)
        self.trustedRetrievers = None

        # User vo paramaters
        self.vo = 'cms'
        self.group = args.get("group", '')
        self.role = args.get("role", '')

        self.logger = args.get("logger", '')

        ## adding ui script to source
        self.uisource = args.get("uisource", '')
        self.cleanEnvironment = args.get("cleanEnvironment", False)

        ## adding credential path
        self.credServerPath = args.get("credServerPath", '/tmp')
        if not self.cmd_exists('voms-proxy-info'):
            raise CredentialException('voms-proxy-info command not found')

    def setEnv(self, cmd):
        """
        Return the source command to be pre added to each command to be executed.
        """
        ret = cmd
        if self.uisource is not None and len(self.uisource) > 0:
            ret = 'source ' + self.uisource + ' && ' + ret
        if self.cleanEnvironment:
            # Need to escape ' in cmd
            ret = ret.replace("'", r"'\''")
            ret = "env -i sh -c '%s'" % ret

        return ret

    def cmd_exists(self, cmd):
        return subprocess.call(self.setEnv("type " + cmd), shell=True,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0
    def getUserCertFilename(self):
        if 'X509_USER_CERT' not in os.environ:
            ucFilename = '~/.globus/usercert.pem'
        else:
            ucFilename = os.environ['X509_USER_CERT']
        return ucFilename

    def getUserCertTimeLeft(self, openSSL=True):
        """
        Return the number of seconds until the expiration of the user cert
        in .globus/usercert.pem or $X509_USER_CERT if set
        Uses openssl by default and fallback to voms-proxy-info in case of problems
        """
        certLocation = self.getUserCertFilename()
        if openSSL:
            out, _, retcode = execute_command('openssl x509 -noout -in %s -dates'
                                              % certLocation, self.logger,
                                              self.commandTimeout)
            if retcode == 0:
                out = out.split('notAfter=')[1]
                if out[-1] == '\n':
                    out = out[:-1]

                possibleFormats = ['%b  %d  %H:%M:%S %Y %Z',
                                   '%b %d %H:%M:%S %Y %Z']
                exptime = None
                for frmt in possibleFormats:
                    try:
                        exptime = datetime.strptime(out, frmt)
                    except ValueError:
                        pass  # try next format
                if not exptime:
                    # If we cannot decode the output in any way print
                    # a message and fallback to voms-proxy-info command
                    self.logger.warning(
                        'Cannot decode "openssl x509 -noout -in %s -dates" date format. Falling back to voms-proxy-info' % certLocation)
                else:
                    # if everything is fine then we are ready to return!!
                    timeleft = (exptime - datetime.utcnow()).total_seconds()
                    return int(timeleft)

        # uses this as a fallback
        timeleft = self.getTimeLeft(proxy=certLocation, checkVomsLife=False)
        if self.retcode:
            raise CredentialException('Cannot get user certificate remaining time with "voms-proxy-info"')

        return timeleft

    def getUserCertEnddate(self, openSSL=True):
        """
        Return the number of days until the expiration of the user cert
        in .globus/usercert.pem or $X509_USER_CERT if set
        Uses openssl by default and fallback to voms-proxy-info in case of problems
        """
        timeleft = self.getUserCertTimeLeft(openSSL)
        daystoexp = int(timeleft // (60. * 60 * 24))

        return daystoexp

    def getProxyDetails(self):
        """
        Return the vo details that should be in the user proxy.
        """
        proxyDetails = "/%s" % self.vo
        if self.group:
            proxyDetails += "/%s" % self.group
        if self.role and self.role != 'NULL':
            proxyDetails += "/Role=%s" % self.role

        return proxyDetails

    def getProxyFilename(self, serverRenewer=False):
        """
        Try to get the proxy file path from:

        1.  the delegated server
        2.  from an ui
        """
        if serverRenewer:
            uniqName = self.userDN + self.vo + self.group + self.role
            proxyFilename = os.path.join(self.credServerPath,
                                         sha1(encodeUnicodeToBytes(uniqName)).hexdigest())
        elif 'X509_USER_PROXY' in os.environ:
            proxyFilename = os.environ['X509_USER_PROXY']
        else:
            proxyFilename = '/tmp/x509up_u' + str(os.getuid())

        return proxyFilename

    def getSubject(self, proxy=None):
        """
        Get proxy subject from a proxy file.
        """
        subject = None

        if proxy is None:
            proxy = self.getProxyFilename()
        getSubjectCmd = "voms-proxy-info -file " + proxy + " -identity"
        subject, _, retcode = execute_command(self.setEnv(getSubjectCmd), self.logger, self.commandTimeout)

        if retcode == 0:
            subject = subject.strip()

        return subject

    def getSubjectFromCert(self, certFile=None):
        """
        Get the subject from cert file.
        """
        subject = ''

        if not certFile:
            certFile = self.getProxyFilename()

        subjFromCertCmd = 'openssl x509 -in ' + certFile + ' -subject -noout'
        subjectResult, _, retcode = execute_command(self.setEnv(subjFromCertCmd), self.logger, self.commandTimeout)

        subject = None
        if retcode == 0:
            subject = subjectResult.split('subject=')[1].strip()

        return subject

    def getUserName(self, proxy=None):
        """
        Get the user name from a proxy file.
        """
        subject = self.getSubject(proxy)
        uName = ''

        for cname in subject.split('/'):
            if cname[:3] == "CN=" and cname[3:].find('proxy') == -1:
                uName = cname[3:]

        return uName

    def getMyproxyUsernameForCRAB(self):
        """
        :return: username (a string) to be used when myproxy-* command is called with the option
               -l | --username        <username> Username for the delegated proxy

        CRAB (and only CRAB as of March 2020) needs to obtain from myproxy a proxy from a credential
        uploaded by another user. So that it can submit jobs and data transfers authenticated
        with the user proxy. For this a special credential is uploaded in myproxy server by
        CRAB Client which is connected to a username associated to the user and has a
        list of authorized retrievers as a list of DN's of CRAB TaskWorkers maintained in
        central CRAB configuration.
        The username is passed to myproxy-* command via the "-l" option
        Three different algorithms for defining this username have been used in CRAB:
        1. the hash of the user DN + the fqdn of the CRAB REST host
        2. the user CERN primary account username + the _CRAB string
        3. the hash of the user DN
        During spring 2020 CRAB migrates from 1. to 2. For a smooth migration the
        new client needs to upload both credentials and the TW will try 2. and fall back to 1.
        Only after all tasks submited with old client are gone from the system, can we change
        to support only 2.
        The reasons to change from 1. to 2. are to make the username readable (helps support) and
        to allow using different REST hosts (helps in K8s world). The reasons for the complicated
        recipe in 1. are unknown, aside some security by obscurity attempt.
        The caller decides if the call to myproxy-* done by this module will use 1. or 2. via
        the userName key in the dictionary passed as argument to Proxy() at __init__ time
           - if the dictionary contains the key 'userName', algorithm 2 is used
           - if the dictionary does not have it, algorithm 1. is used
        But at times user change their DN and can't act on credentials stored in myproxy
        with old DN. They need to switch to a new credential name, for this 3. neede3d to be put back
        in Summer 2020.
        """
        if self.userName:
            self.logger.debug("using %s as credential login name", self.userName)
            username = self.userName
        else:
            if self.myproxyAccount:
                self.logger.debug(
                    "Calculating hash of %s for credential name" % (self.userDN + "_" + self.myproxyAccount))
                username = sha1(encodeUnicodeToBytes(self.userDN + "_" + self.myproxyAccount)).hexdigest()
            else:
                self.logger.debug(
                    "Calculating hash of %s for credential name" % (self.userDN))
                username = sha1(encodeUnicodeToBytes(self.userDN)).hexdigest()
        return username

    def checkAttribute(self, proxy=None):
        """
        Check attributes from a proxy file.
        """
        valid = True

        if proxy is None:
            proxy = self.getProxyFilename()

        checkAttCmd = 'voms-proxy-info -fqan -file ' + proxy
        proxyDetails = self.getProxyDetails()

        attribute = execute_command(self.setEnv(checkAttCmd), self.logger, self.commandTimeout)

        if not re.compile(r"^" + proxyDetails).search(attribute[0]):
            valid = False

        return valid

    def create(self):
        """
        Proxy creation.
        """
        createCmd = 'voms-proxy-init -voms %s:%s -valid %s %s' % (
            self.vo, self.getProxyDetails(), self.proxyValidity, '-rfc' if self.rfcCompliant else '')
        execute_command(self.setEnv(createCmd), self.logger, self.commandTimeout, redirect=False)

        return

    def renew(self):
        """
        Proxy renew.
        """
        self.create()
        return

    def destroy(self, credential=None):
        """
        Proxy destruction.
        """
        if not credential:
            credential = self.getProxyFilename()

        destroyCmd = 'rm -f %s' % credential
        execute_command(destroyCmd, self.logger, self.commandTimeout)

        return

    def delegate(self, credential=None, serverRenewer=False, nokey=False):
        """
        Delegate the user proxy to myproxy.
        It is possible also to delegate a server
        (specifying serverRenewer = True) to
        manage your proxy in myproxy server.
        """
        if not credential:
            credential = self.getProxyFilename(serverRenewer)

        if self.myproxyServer:
            myproxyDelegCmd = 'export GT_PROXY_MODE=%s ; myproxy-init -d -n -s %s' % (
                'rfc' if self.rfcCompliant else 'old', self.myproxyServer)

            if nokey is True:
                myproxyUsername = self.getMyproxyUsernameForCRAB()
                myproxyDelegCmd = 'export GT_PROXY_MODE=%s ; myproxy-init -d -n -s %s -x -R \'%s\' -x -Z \'%s\' -l \'%s\' -t 168:00 -c %s' \
                                  % ('rfc' if self.rfcCompliant else 'old', self.myproxyServer, self.serverDN, \
                                     self.serverDN, myproxyUsername, self.myproxyValidity)
            elif serverRenewer and len(self.serverDN.strip()) > 0:
                serverCredName = sha1(encodeUnicodeToBytes(self.serverDN)).hexdigest()
                myproxyDelegCmd += ' -x -R \'%s\' -Z \'%s\' -k %s -t 168:00 -c %s ' \
                                   % (self.serverDN, self.serverDN, serverCredName, self.myproxyValidity)
            _, stderr, _ = execute_command(self.setEnv(myproxyDelegCmd), self.logger, self.commandTimeout)
            if stderr.find('proxy will expire') > -1:
                raise CredentialException('Your certificate is shorter than %s ' % self.myproxyValidity)
        else:
            self.logger.error("myproxy server not set for the proxy %s" % credential)

        return

    def getMyProxyTimeLeft(self, proxy=None, serverRenewer=False, nokey=False):
        """
        Get myproxy timeleft. Speciying serverRenewer=True means
        that your are delegating your proxy management in myproxy
        to a server.
        """
        proxyTimeleft = -1
        if self.myproxyServer:
            if nokey is True and serverRenewer is True:
                myproxyUsername = self.getMyproxyUsernameForCRAB()
                checkMyProxyCmd = 'myproxy-info -l %s -s %s' % (myproxyUsername, self.myproxyServer)
                output, _, retcode = execute_command(self.setEnv(checkMyProxyCmd), self.logger, self.commandTimeout)
                if retcode > 0 or not output:
                    return proxyTimeleft

                trustedRetrList = re.compile('trusted retrieval policy: (.*)').findall(output)
                if len(trustedRetrList) > 1 or len(trustedRetrList) == 0:
                    raise CredentialException(
                        "Unexpected result while decoding trusted retrievers list: " + str(trustedRetrList))
                else:
                    self.trustedRetrievers = trustedRetrList[0]

                timeleftList = re.compile(
                    "timeleft: (?P<hours>[\\d]*):(?P<minutes>[\\d]*):(?P<seconds>[\\d]*)").findall(output)
                if len(timeleftList) > 1 or len(timeleftList) == 0:
                    raise CredentialException(str(timeleftList))
                else:
                    hours, minutes, seconds = timeleftList[0]
                    proxyTimeleft = int(hours) * 3600 + int(minutes) * 60 + int(seconds)
                    return proxyTimeleft

            if not proxy:
                proxy = self.getProxyFilename(serverRenewer)
            checkMyProxyCmd = 'myproxy-info -d -s ' + self.myproxyServer
            output, _, retcode = execute_command(self.setEnv(checkMyProxyCmd), self.logger, self.commandTimeout)

            if retcode > 0 or not output:
                return proxyTimeleft

            timeleftList = re.compile("timeleft: (?P<hours>[\\d]*):(?P<minutes>[\\d]*):(?P<seconds>[\\d]*)").findall(
                output)
            hours, minutes, seconds = 0, 0, 0

            if not serverRenewer:

                try:
                    hours, minutes, seconds = timeleftList[0]
                    proxyTimeleft = int(hours) * 3600 + int( \
                        minutes) * 60 + int(seconds)
                except Exception as e:
                    self.logger.error('Error extracting timeleft from proxy %s' % str(e))

            elif len(self.serverDN.strip()) > 0:
                serverCredName = sha1(encodeUnicodeToBytes(self.serverDN)).hexdigest()
                credNameList = re.compile(" name: (?P<CN>.*)").findall(output)

                if len(timeleftList) == len(credNameList):
                    credTimeleftList = timeleftList
                else:
                    credTimeleftList = timeleftList[1:]

                if serverCredName not in credNameList:
                    self.logger.error('Your proxy needs retrieval and renewal policies for the requested server.')
                    proxyTimeleft = 0
                else:
                    try:
                        hours, minutes, seconds = credTimeleftList[credNameList.index(serverCredName)]
                        proxyTimeleft = int(hours) * 3600 + int(minutes) * 60 + int(seconds)
                    except Exception as e:
                        self.logger.error('Error extracting timeleft from credential name %s' % str(e))

            else:
                self.logger.error('Configuration Error')

        else:
            self.logger.error("myproxy server not set")

        return proxyTimeleft

    def checkMyProxy(self, proxy=None, checkRenewer=False):
        """
        Return True if myproxy validity is bigger than minTime.
        """
        if self.myproxyServer:
            valid = True

            if not proxy:
                proxy = self.getProxyFilename(checkRenewer)

            checkMyProxyCmd = 'myproxy-info -d -s ' + self.myproxyServer
            output, _, retcode = execute_command(self.setEnv(checkMyProxyCmd), self.logger, self.commandTimeout)

            if retcode > 0 and not output:
                valid = False
                return valid

            minTime = self.myproxyMinTime * 24 * 3600
            # regex to extract the right information
            timeleftList = re.compile("timeleft: (?P<hours>[\\d]*):(?P<minutes>[\\d]*):(?P<seconds>[\\d]*)").findall(
                output)
            timeleft, hours, minutes, seconds = 0, 0, 0, 0

            # the first time refers to the flat user proxy,
            # the other ones are related to the server credential name
            if not checkRenewer:

                try:
                    hours, minutes, seconds = timeleftList[0]
                    timeleft = int(hours) * 3600 + int(minutes) * 60 + int(seconds)
                except Exception as e:
                    self.logger.error('Error extracting timeleft from proxy %s' % str(e))
                    return False

                if timeleft < minTime:
                    self.logger.error('Your proxy will expire in:\n\t%s hours %s minutes %s seconds\n the minTime : %s'
                                      % (hours, minutes, seconds, minTime))
                    valid = False

            # check the timeleft for the required server
            elif len(self.serverDN.strip()) > 0:

                serverCredName = sha1(encodeUnicodeToBytes(self.serverDN)).hexdigest()
                credNameList = re.compile(" name: (?P<CN>.*)").findall(output)

                # check if the server credential exists
                if serverCredName not in credNameList:
                    self.logger.error('Your proxy needs retrieval and renewal policies for the requested server.')
                    return False

                if len(timeleftList) == len(credNameList):
                    credTimeleftList = timeleftList
                else:
                    credTimeleftList = timeleftList[1:]

                # clean up expired credentials for other servers anyway
                destroyListCred(credNameList, credTimeleftList, self.logger, self.commandTimeout)

                try:
                    hours, minutes, seconds = credTimeleftList[credNameList.index(serverCredName)]
                    timeleft = int(hours) * 3600 + int(minutes) * 60 + int(seconds)
                except Exception as e:
                    self.logger.error('Error extracting timeleft from credential name %s' % str(e))
                    return False

                if timeleft < minTime:
                    logMsg = 'Your credential for the required server will expire in:\n\t%s hours %s minutes %s seconds\n' \
                             % (hours, minutes, seconds)
                    self.logger.error(logMsg)
                    valid = False

            else:
                self.logger.error('Configuration Error')
                valid = False

        else:
            self.logger.error('Error delegating credentials : myproxyserver is not specified.')
            valid = False

        return valid

    def logonRenewMyProxy(self, proxyFilename=None, credServerName=None):
        """
        Refresh/retrieve proxyFilename in/from myproxy.
        """
        if not proxyFilename:
            proxyFilename = self.getProxyFilename(serverRenewer=True)

        attribute = self.getAttributeFromProxy(proxyFilename)
        if not attribute:
            attribute = self.getProxyDetails()
        voAttribute = self.prepareAttForVomsRenewal(attribute)

        # compose the delegation or renewal commands
        # with the regeneration of Voms extensions
        cmdList = []
        cmdList.append('unset X509_USER_CERT X509_USER_KEY')
        cmdList.append('&& env')
        cmdList.append('X509_USER_CERT=%s' % self.serverCert)
        cmdList.append('X509_USER_KEY=%s' % self.serverKey)

        ## get a new delegated proxy
        uniqName = self.userDN + self.vo + self.group + self.role
        proxyFilename = os.path.join(self.credServerPath,
                                     sha1(encodeUnicodeToBytes(uniqName)).hexdigest())
        # Note that this is saved in a temporary file with the pid appended to the filename. This way we will avoid adding many
        # signatures later on with vomsExtensionRenewal in case of multiple processing running at the same time
        tmpProxyFilename = proxyFilename + '.' + str(os.getpid())

        myproxyUsername = self.getMyproxyUsernameForCRAB()

        cmdList.append('myproxy-logon -d -n -s %s -o %s -l \"%s\" -t 168:00'
                       % (self.myproxyServer, tmpProxyFilename, myproxyUsername) )
        logonCmd = ' '.join(cmdList)
        msg, _, retcode = execute_command(self.setEnv(logonCmd), self.logger, self.commandTimeout)

        if retcode > 0:
            self.logger.error("Unable to retrieve delegated proxy using login %s for user DN %s! Exit code:%s output:%s",
                              myproxyUsername, self.userDN, retcode, msg)
            return proxyFilename

        self.vomsExtensionRenewal(tmpProxyFilename, voAttribute)
        os.rename(tmpProxyFilename, proxyFilename)

        return proxyFilename

    def prepareAttForVomsRenewal(self, attribute='/cms'):
        """
        Prepare attribute for the voms renewal.
        """
        # prepare the attributes for voms extension
        voAttribute = self.vo + ':' + attribute

        # Clean attribute to extend voms
        voAttribute = voAttribute.replace('/Role=NULL', '')
        voAttribute = voAttribute.replace('/Capability=NULL', '')

        return voAttribute

    def vomsExtensionRenewal(self, proxy, voAttribute='cms'):
        """
        Renew voms extension of the proxy
        """
        # get RFC/noRFC type for retrieved flat proxy
        msg, _, retcode = execute_command(self.setEnv('voms-proxy-info -type -file %s' % proxy), self.logger,
                                          self.commandTimeout)
        if retcode > 0:
            self.logger.error('Cannot get proxy type %s' % msg)
            return
        isRFC = msg.startswith('RFC')  # can be 'RFC3820 compliant impersonation proxy' or 'RFC compliant proxy'

        # get validity time for retrieved flat proxy
        cmd = 'grid-proxy-info -file ' + proxy + ' -timeleft'
        timeLeft, _, retcode = execute_command(self.setEnv(cmd), self.logger, self.commandTimeout)

        if retcode != 0:
            self.logger.error("Error while checking retrieved proxy timeleft for %s" % proxy)
            return


        # timeLeft indicates how many seconds the proxy is still valid for.
        # We will add a VOMS extension via voms-proxy-init -noregen but if we
        # ask for an exactly matching validity time, we often end with
        #  Warning: your certificate and proxy will expire Tue Mar 31 23:46:06 2020
        #  which is within the requested lifetime of the proxy
        # which causes a non zero exit status and hence a false error logging.
        # Therefore let's take 10 minutes off
        vomsTime = int(timeLeft.strip()) - 600

        vomsValid = '00:00'
        if vomsTime > 0:
            vomsValid = "%d:%02d" % (int(vomsTime // 3600), int((vomsTime % 3600) // 60))
        self.logger.debug('Requested voms validity: %s' % vomsValid)

        ## set environ and add voms extensions
        cmdList = []
        cmdList.append('env')
        cmdList.append('X509_USER_PROXY=%s' % proxy)
        cmdList.append('voms-proxy-init -noregen -voms %s -out %s -bits 2048 -valid %s %s'
                       % (voAttribute, proxy, vomsValid, '-rfc' if isRFC  else ''))
        cmd = ' '.join(cmdList)
        msg, _, retcode = execute_command(self.setEnv(cmd), self.logger, self.commandTimeout)

        if retcode > 0:
            self.logger.error('Unable to renew proxy voms extension: %s' % msg)

        return

    def renewMyProxy(self, proxy=None, serverRenewer=False):
        """
        Renew MyProxy
        """
        if not proxy:
            proxy = self.getProxyFilename(serverRenewer)
        self.delegate(proxy, serverRenewer)

        return

    ##################### Check timeleft
    def getTimeLeft(self, proxy=None, checkVomsLife=True):
        """
        Get proxy timeleft. Validate the proxy timeleft
        with the voms life.
        """
        timeLeft = 0
        if not proxy:
            proxy = self.getProxyFilename()

        timeLeftCmd = 'voms-proxy-info -file ' + proxy + ' -timeleft'
        timeLeftLocal, _, self.retcode = execute_command(self.setEnv(timeLeftCmd), self.logger, self.commandTimeout)

        if self.retcode != 0:
            self.logger.error("Error while checking proxy timeleft for %s" % proxy)
            return timeLeft
        try:
            timeLeft = int(timeLeftLocal.strip())
        except ValueError:
            timeLeft = sum(int(x) * 60 ** i for i, x in enumerate(reversed(timeLeftLocal.strip().split(":"))))

        if checkVomsLife and timeLeft > 0:
            ACTimeLeftLocal = self.getVomsLife(proxy)
            if timeLeft != ACTimeLeftLocal:
                msg = "Proxy lifetime %s secs is different from " % timeLeft
                msg += "voms extension lifetime %s secs for proxy: %s" % (ACTimeLeftLocal, proxy)
                self.logger.debug(msg)
            timeLeft = min(timeLeft, ACTimeLeftLocal)

        return timeLeft

    def getVomsLife(self, proxy):
        """
        Get proxy voms life.
        """
        result = 0
        cmd = 'voms-proxy-info -file ' + proxy + ' -actimeleft'
        ACtimeLeftLocal, _, retcode = execute_command(self.setEnv(cmd), self.logger, self.commandTimeout)

        if retcode != 0:
            return result
        try:

            result = int(ACtimeLeftLocal)
        except ValueError:
            result = sum(int(x) * 60 ** i for i, x in enumerate(reversed(ACtimeLeftLocal.split(":"))))

        return result

    def getAttributeFromProxy(self, proxy, allAttributes=False):
        """
        Get proxy attribute.
        Build the proxy attribute from existing and not from parameters as
        done by getProxyDetails.
        """
        roleCapCmd = 'env X509_USER_PROXY=%s voms-proxy-info -fqan' % proxy
        attribute, _, retcode = execute_command(self.setEnv(roleCapCmd),
                                                self.logger,
                                                self.commandTimeout)
        if retcode == 0:
            if allAttributes:
                return list(filter(bool, attribute.split('\n')))
            else:
                return attribute.split('\n')[0]
        else:
            return ''

    def getUserGroupAndRoleFromProxy(self, proxy):
        """
        Get user group and role from the proxy attribute.
        """
        group, role = '', ''
        attribute = self.getAttributeFromProxy(proxy)
        if attribute:
            attributeToList = attribute.split('/')
            if len(attributeToList) > 4:
                group = attributeToList[2]
                role = attributeToList[3].split('=')[1]
            else:
                role = attributeToList[2].split('=')[1]

        return group, role

    def getAllUserGroups(self, proxy):
        """
        Get all the attributes for the user using getAttributeFromProxy
        and strip the ROLE and CAPABILITIES part.

        Return a generator of things like '/cms/integration', '/cms'
        """
        attributes = self.getAttributeFromProxy(proxy, allAttributes=True)
        for attribute in attributes:
            splAttr = attribute.split('/')  # splitted attribut
            filtAttr = [part for part in splAttr if
                        not (part.startswith('Role=') or part.startswith('Capability='))]  # filtered attribute
            yield '/'.join(filtAttr)



#!/usr/bin/env python
"""
_SetStatus_

MySQL implementation for altering job status
"""


from WMCore.Database.DBFormatter import DBFormatter

class SetStatus(DBFormatter):
    """
    _SetStatus_

    Set the status of a list of jobs
    """


    sql = """UPDATE bl_runjob SET sched_status =
               (SELECT id FROM bl_status WHERE name = :status)
               WHERE bl_runjob.id = :id"""


    def execute(self, jobs, status, conn = None, transaction = False):
        """
        _execute_

        Changes the status of a list of jobs to the given status
        Expects a list of IDs
        """

        if len(jobs) == 0:
            return

        binds = []
        for jobid in jobs:
            binds.append({'id': jobid, 'status': status})

        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
MySQL implementation for retrieving the total number of jobs
for a given workflow in a given wmbs status
"""
from __future__ import print_function, division

from WMCore.Database.DBFormatter import DBFormatter


class GetNumberOfJobsForWorkflowStatus(DBFormatter):
    """
    Get the number of jobs for a workflow/task combination
    in a given job status.
    """

    sql = """SELECT COUNT(*) FROM wmbs_job
             INNER JOIN wmbs_jobgroup ON wmbs_job.jobgroup = wmbs_jobgroup.id
             INNER JOIN wmbs_subscription ON wmbs_subscription.id = wmbs_jobgroup.subscription
             INNER JOIN wmbs_workflow ON wmbs_subscription.workflow = wmbs_workflow.id
             WHERE wmbs_workflow.name = :workflow AND
                   wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = :status)
               """

    def execute(self, workflow, status='executing', conn=None, transaction=False):
        """
        Execute the SQL and return the number of jobs

        :param workflow: string with the workflow name
        :param status: string with the wmbs state
        :param conn: database connection object
        :param transaction: boolean to use or not a transaction
        :return: integer with the total number of jobs
        """
        binds = {'workflow': workflow, 'status': status}
        result = self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)

        return self.format(result)[0][0]



#!/usr/bin/env python

"""
_WMRuntimeMonitor_

This is the base class for monitors
"""




from builtins import object
import threading
import os

from WMCore.WMException                         import WMException
from WMCore.WMSpec.Steps.Executor               import getStepSpace
from WMCore.WMSpec.WMStep                       import WMStepHelper

getStepName = lambda step: WMStepHelper(step).name()

class WMRuntimeMonitorException(WMException):
    """
    _StepFactortyException_

    It's like an exception class that does nothing

    """
    pass

class WMRuntimeMonitor(object):


    def __init__(self):
        self.currentStep      = None
        self.currentStepName  = None
        self.currentStepSpace = None
        self.job              = None
        self.task             = None
        return



    def initMonitor(self, task, job, logPath, args = {}):
        return


    def shutdown(self):
        """
        Shutdown method, will be called before object is deleted
        at end of job.
        """
        pass


    def periodicUpdate(self):
        """
        Periodic update.
        """
        pass


    def jobStart(self, task):
        """
        Job start notifier.
        """
        pass


    def stepStart(self, step):
        """
        Tasked started notifier.
        """
        self.currentStep      = step
        self.currentStepName  = getStepName(step)
        self.currentStepSpace = getStepSpace(self.currentStepName)
        return


    def stepEnd(self, step, stepReport):
        """
        Tasked ended notifier.
        """
        self.currentStep      = None
        self.currentStepName  = None
        self.currentStepSpace = None
        return



    def jobEnd(self, task):
        """
        Job ended notifier.
        """
        pass

    def jobKilled(self):
        """
        Job killed notifier.
        """
        pass

    def stepKilled(self):
        """
        Task killed notifier.
        """
        pass

    def killCMSRun(self):
        """
        Kills all cmsRun processes owned by user

        """

        uid = os.getuid()



#!/usr/bin/env python
"""
_GetParentIDsByID_

Oracle implementation of GetParentIDsByID
"""




from WMCore.WMBS.MySQL.Files.GetParentIDsByID import GetParentIDsByID \
     as GetParentIDsMySQL

class GetParentIDsByID(GetParentIDsMySQL):
    pass
